/**
	Copyright (C) 2006  Volker Wiendl

    This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License version 2 as published 
	by the Free Software Foundation.
    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
	without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
	See the GNU General Public License for more details.
    You should have received a copy of the GNU General Public License along with this program; 
	if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA
*/
#ifndef HTPCCOMMAND_H_
#define HTPCCOMMAND_H_

#include <Qt/qvariant.h>

class OutputInterface;
/**
 * \brief Class for handling different Htpc Commands transfered between InputInterface Plugins and OutputInterface Plugins
 */
class HtpcCommand
{
public:

		/**
		 * Basic Constructor
		 */
		HtpcCommand() : m_id(-1), m_name("not assigned"), m_data(), m_plugin(0) {}
		/**
		 * Constructor
		 * @param plugin the output interface this HtpcCommand will belong to
		 * @param id unique id for this command within the OutputInterface plugin
		 * @param name of this command (used for configuration dialogs)
		 * @param data optional data for this command
		 */
		HtpcCommand(OutputInterface* plugin, int id, const QString& name, QVariant data = QVariant()) 
		{ 
			m_plugin = plugin;
			m_id = id; 
			m_name = name;
			m_data = data;
		}
		virtual ~HtpcCommand() {}

		/**
		 * \brief gets the name of the command
		 * @return const QString& reference to the name of the command
		 */
		const QString& getName() const {return m_name;}

		/**
		 * \brief gets the id of the command
		 * @return const int the id of the command
		 */
		const int& getID() const {return m_id;}

		/**
		 * \brief gets the data associated with this command
		 * @return const QVariant& reference to the data associated with this command
		 */
		const QVariant& getData() const {return m_data;}
		
		/**
		 * Sets the data of this command
		 * @param data a QVariant containing data for the output plugin
		 */
		void setData(const QVariant& data) {m_data = data;}

		/**
		 * Returns the plugin this command belongs to 
		 * @return OutputInterface* the plugin this command was by
		 */
		OutputInterface* const getPlugIn() const {return m_plugin;}


		/**
		 * A HtpcControl is equal to another one if the id and the name is equal
		 */
		bool operator==(const HtpcCommand& command) {
			return getName()==command.getName() && getID() == command.getID() && getPlugIn() == command.getPlugIn();
		}

protected:

		int		 m_id;
		QString	 m_name;
		QVariant m_data;
		OutputInterface* m_plugin;

};
Q_DECLARE_METATYPE(HtpcCommand)
#endif
