/**
	Copyright (C) 2006  Volker Wiendl

    This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License version 2 as published 
	by the Free Software Foundation.
    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
	without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
	See the GNU General Public License for more details.
    You should have received a copy of the GNU General Public License along with this program; 
	if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA
*/
#ifndef INPUTINTERFACE_H_
#define INPUTINTERFACE_H_

#include <Qt/qdom.h>
#include <Qt/qplugin.h>
#include <Qt/qthread.h>
#include "../HTPCControl/HTPCCommand.h"
#include "../HTPCControl/HtpcInterface.h"

class OutputInterface;

/**
 * \brief Interface for plugins creating input for output plugins
 * 
 * This interface is the base class for all plugins creating any data that can be used
 * by an output plugin. 
 * e.g. a remote control plugin would be an input plugin for a winamp output plugin that
 * take the "input" key to create an "output" action (e.g. next track)
 */
class InputInterface : public HtpcInterface, public QThread
{	

    public:
        virtual ~InputInterface() {}

		/**
		 * This method has to be implemented by subclasses to connect the input interface to a given
		 * output interface. The configuration of the connection will be read from the QDomElement instance
		 * 
		 * @param output the output interface the plugin should be connected to
		 * @param configuration an XML configuration that holds the information about the different HtpcCommand settings
		 * for this connection
		 * @return false in case of an error
		 */
		virtual bool connectPlugIn(OutputInterface* output, const QDomElement& configuration) = 0;

		/**
		 * This method has to be implemented in subclasses. It is supposed to show a configuration dialog
		 * where the connection options for the given output interface will be shown
		 * 
		 * @param output instance of an output interface the current input plugin should connect to
		 * @param configuration reference to the XML configuration node for this input plugin (currently not used by any plugin)
		 * @return false in case of an error
		 */
		virtual bool configPlugInConnection(OutputInterface* output, QDomElement& configuration) = 0;
		
		/**
		 * This method has to be implemented in subclassed. The plugin should save it's current connections into the given
		 * QDomDocument instance.
		 * 
		 * @param doc reference to the XML configuration node for this plugin
		 */
		virtual bool saveConfig(QDomDocument& doc) = 0;

		/**
		 * Since InputInterface is inherited from QThread this method has to be implemented in subclasses
		 */
		virtual void run() = 0;

		/**
		 * Has to be implemented in subclasses. The method is supposed to stop the thread and exit it's mainloop.
		 */
		virtual void stop() = 0;

};

Q_DECLARE_INTERFACE(InputInterface,
                        "HTPCControl.InputInterface/1.0")
#endif
