#include "PlugInDialog.h"

#include "InputInterface.h"
#include "OutputInterface.h"
#include "PlugInListWidgetItem.h"
#include <Qt/qmessagebox.h>

PlugInDialog::PlugInDialog(QWidget* parent /*=0*/, Qt::WFlags flags /* = 0 */) : QDialog(parent, flags)
{
	setupUi(this);
	m_configureInputButton->setEnabled(false);
	m_configureOutputButton->setEnabled(false);
	m_connectButton->setEnabled(false);	
	connect(m_inputPlugIns, SIGNAL(currentItemChanged ( QListWidgetItem*, QListWidgetItem*)), this, SLOT(inputPlugInSelected(QListWidgetItem*, QListWidgetItem*)));
	connect(m_outputPlugIns, SIGNAL(currentItemChanged ( QListWidgetItem*, QListWidgetItem*)), this, SLOT(outputPlugInSelected(QListWidgetItem*, QListWidgetItem*)));
	connect(m_connectButton, SIGNAL(clicked(bool)), this, SLOT(connectButtonPressed()));
	connect(m_configureInputButton, SIGNAL(clicked()), this, SLOT(configureInputPlugIn()));
	connect(m_configureOutputButton, SIGNAL(clicked()), this, SLOT(configureOutputPlugIn()));
}

PlugInDialog::~PlugInDialog()
{
}


void PlugInDialog::init(const QFileInfo& configFile, const QList<InputInterface*>& inputPlugIns, const QList<OutputInterface*>& outputPlugIns)
{
	m_configFile.setFileName(configFile.absoluteFilePath());	
	if (!m_configFile.open(QIODevice::ReadOnly))
	{
		QMessageBox::warning(this, tr("Error"), tr("Couldn't open file: \n")+configFile.absoluteFilePath() );
		return;
	}
	m_doc = QDomDocument("Configuration");
	if (!m_doc.setContent(&m_configFile))
	{
		QMessageBox::warning(this, tr("Error"), tr("Error reading DOM structure from file: \n")+configFile.absoluteFilePath());
		m_configFile.close();
		//return;
	}
	m_configFile.close();	
	QList<InputInterface*>::const_iterator inIter = inputPlugIns.constBegin();
	while(inIter!=inputPlugIns.constEnd())
	{
		PlugInListWidgetItem* item = new PlugInListWidgetItem((*inIter)->getPlugInName(), m_inputPlugIns);
		item->setInterface(*inIter);
		++inIter;
	}
	QList<OutputInterface*>::const_iterator outIter = outputPlugIns.constBegin();
	while(outIter!=outputPlugIns.constEnd())
	{
		PlugInListWidgetItem* item = new PlugInListWidgetItem((*outIter)->getPlugInName(), m_outputPlugIns);
		item->setInterface(*outIter);
		++outIter;
	}	
}

void PlugInDialog::connectButtonPressed()
{
	InputInterface* inputPlugIn = (InputInterface*) ((PlugInListWidgetItem*) m_inputPlugIns->currentItem())->getInterface();
	OutputInterface* outputPlugIn = (OutputInterface*)((PlugInListWidgetItem*) m_outputPlugIns->currentItem())->getInterface();
	QDomElement docElem = m_doc.documentElement();
	const QDomNodeList& inputInterfaceNodes = docElem.elementsByTagName("InputInterface");
	for (unsigned int i=0; i<inputInterfaceNodes.length(); ++i)
	{
		QDomElement inputInterfaceNode = inputInterfaceNodes.item(i).toElement();
		QString inputInterfaceName = inputInterfaceNode.attribute("name");
		if (m_inputPlugIns->currentItem()->text()==inputInterfaceName)
		{			
			if (inputPlugIn->configPlugInConnection(outputPlugIn, inputInterfaceNode))
			{
				// don't save configuration yet
				// saving will be done by the PlugInManager if the Ok button of the PlugInDialog was pressed 
			}
			return;
		}
	}
	QDomElement inputInterface = docElem.toDocument().createElement("InputInterface");
	inputInterface.setAttribute("name", inputPlugIn->getPlugInName());
	inputInterface.setAttribute("priority", inputPlugIn->priority());
	if (inputPlugIn->configPlugInConnection(outputPlugIn, inputInterface))
		docElem.appendChild(inputInterface);
}

void PlugInDialog::inputPlugInSelected(QListWidgetItem *current, QListWidgetItem *previous)
{
	m_configureInputButton->setEnabled(current!=0);	
	m_connectButton->setEnabled(current && m_outputPlugIns->currentItem());
	if (current)
		m_inputPlugIns->selectionModel()->select(m_inputPlugIns->currentIndex(), QItemSelectionModel::Select);
}

void PlugInDialog::outputPlugInSelected(QListWidgetItem *current, QListWidgetItem *previous)
{
	m_configureOutputButton->setEnabled(current!=0);	
	m_connectButton->setEnabled(current && m_inputPlugIns->currentItem());
	if (current)
		m_outputPlugIns->selectionModel()->select(m_outputPlugIns->currentIndex(),QItemSelectionModel::Select);
}

void PlugInDialog::configureInputPlugIn()
{
	((PlugInListWidgetItem*)m_inputPlugIns->currentItem())->getInterface()->configure();
}

void PlugInDialog::configureOutputPlugIn()
{
	((PlugInListWidgetItem*)m_outputPlugIns->currentItem())->getInterface()->configure();
}
