/**
	Copyright (C) 2006  Volker Wiendl

    This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License version 2 as published 
	by the Free Software Foundation.
    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
	without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
	See the GNU General Public License for more details.
    You should have received a copy of the GNU General Public License along with this program; 
	if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA
*/

#include "PlugInManager.h"
#include "InputInterface.h"
#include "OutputInterface.h"

#include <Qt/qdir.h>
#include <Qt/qapplication.h>
#include <Qt/qpluginloader.h>
#include <Qt/qvariant.h>
#include <Qt/qtextstream.h>
#include <Qt/qfile.h>

PlugInManager::PlugInManager(QObject* parent /*=0*/) : QObject(parent)
{
	QDir pluginsDir = QDir(qApp->applicationDirPath());
	pluginsDir.cd("plugins");
	QStringList pluginFileNames;	
	foreach (QString fileName, pluginsDir.entryList(QDir::Files)) {
		if (fileName.endsWith(QString(".dll")))
		{
			QPluginLoader loader(pluginsDir.absoluteFilePath(fileName));
			QObject *plugin = loader.instance();
			if (plugin) {
				plugin->setParent(this);
				InputInterface *controlPlugIn = qobject_cast<InputInterface *>(plugin);
				if (controlPlugIn)
					m_input_plug_ins.push_back(controlPlugIn);										
				OutputInterface* outputPlugIn = qobject_cast<OutputInterface*>(plugin);
				if (outputPlugIn)
					m_output_plug_ins.push_back(outputPlugIn);
				// ...
			}			
		}
	}	
}

PlugInManager::~PlugInManager()
{
	InputInterfaceList::iterator inIter = m_input_plug_ins.begin();
	while (inIter!=m_input_plug_ins.end())
	{			
		(*inIter)->stop();
		++inIter;
	}			
}

bool PlugInManager::init(const QString& fileName)
{
	QFile configFile(fileName);
	if (!configFile.open(QIODevice::ReadOnly))
		return false;
	QDomDocument doc("Configuration");
	if (!doc.setContent(&configFile))
	{
		configFile.close();
		return false;
	}
	configFile.close();
	
	QDomElement docElem = doc.documentElement();
	const QDomNodeList& inputInterfaces = docElem.elementsByTagName("InputInterface");
	for (unsigned int i=0; i<inputInterfaces.length(); ++i)
	{
		QDomElement inputInterface = inputInterfaces.item(i).toElement();
		QString inputInterfaceName = inputInterface.attribute("name");
		int priority = inputInterface.attribute("priority", "1").toInt();
		InputInterfaceList::iterator inIter = m_input_plug_ins.begin();
		while (inIter!=m_input_plug_ins.end())
		{
			if (inputInterfaceName.toLower() == (*inIter)->getPlugInName().toLower())
			{
				const QDomNodeList& outputInterfaces = inputInterface.toElement().elementsByTagName("OutputInterface");				
				for (unsigned int j=0; j<outputInterfaces.length(); ++j)
				{
					QDomNode outputInterface = outputInterfaces.item(j);
					QString outputInterfaceName = outputInterface.toElement().attribute("name");
					OutputInterfaceList::const_iterator outIter = m_output_plug_ins.constBegin();
					while (outIter!=m_output_plug_ins.end())
					{
						if ((*outIter)->getPlugInName().compare(outputInterfaceName, Qt::CaseInsensitive)==0)
							(*inIter)->connectPlugIn(*outIter, outputInterface.toElement());
						++outIter;
					}
				}			
				(*inIter)->start(QThread::Priority(priority));
			}
			++inIter;
		}
	}
	return true;
}
	

const QList<InputInterface*> PlugInManager::getInputPlugIns()
{
	return m_input_plug_ins;
}
const QList<OutputInterface*> PlugInManager::getOutputPlugIns()
{
	return m_output_plug_ins;
}

void PlugInManager::save(const QString& fileName)
{
	if (m_input_plug_ins.size() > 0 && m_output_plug_ins.size()>0)
	{
		QFile configFile(fileName, this);	
		if (configFile.open(QFile::WriteOnly | QFile::Truncate)) 
		{
			QTextStream out(&configFile);
			QDomDocument doc("Configuration");		
			QDomElement rootNode = doc.createElement("Configuration");
			doc.appendChild(rootNode);
			InputInterfaceList::iterator inIter = m_input_plug_ins.begin();
			while (inIter!=m_input_plug_ins.end())
			{			
				(*inIter)->saveConfig(doc);
				++inIter;
			}			
			doc.save(out, 6);
			configFile.close();
		}
	}
}
