#include "ConnectionDialog.h"

#include <qt/qmessagebox.h>

#include "../HTPCControl/OutputInterface.h"
#include "ircodeevent.h"

ConnectionDialog::ConnectionDialog(QWidget* parent /*=0*/, Qt::WFlags wflags /*= 0*/) : QDialog(parent, wflags), m_learning(false)
{
	setupUi(this);
	connect(m_irCodeTree, SIGNAL(currentItemChanged(QTreeWidgetItem*, QTreeWidgetItem*)), this, SLOT(itemSelected()));
	connect(m_learnCodeButton, SIGNAL(clicked()), this, SLOT(learnButtonPressed()));
	connect(m_outputCommands, SIGNAL(itemSelectionChanged()), this, SLOT(itemSelected()));
	connect(m_addCommandButton, SIGNAL(clicked()), this, SLOT(commandAssociated()));
	connect(m_antiRepeat, SIGNAL(valueChanged(int)), this, SLOT(updateIrCode(int)));
}

ConnectionDialog::~ConnectionDialog(void)
{
}

void ConnectionDialog::init(const QMap<IRCode,QList<HtpcCommand>*>& commandMap, OutputInterface *outputPlugIn)
{
	m_commandMap = commandMap;
	QList<HtpcCommand>::const_iterator iter;
	QMap<IRCode, QList<HtpcCommand>*>::const_iterator codeIter = m_commandMap.constBegin();
	// insert existing connections
	while (codeIter != m_commandMap.constEnd())
	{				
		if (codeIter.value())
		{
			QTreeWidgetItem* irItem = 0;
			QList<QTreeWidgetItem*>& irCode = m_irCodeTree->findItems(QString::number(codeIter.key().code), Qt::MatchExactly);
			if (irCode.size()==0)
			{
				irItem = new QTreeWidgetItem(m_irCodeTree, QStringList(QString::number(codeIter.key().code)));
				irItem->setData(0, Qt::UserRole, QVariant::fromValue(codeIter.key()));
			}
			else
				irItem = irCode.at(0);				
			iter = codeIter.value()->constBegin();
			while(iter!=codeIter.value()->constEnd())
			{		
				if (iter->getPlugIn()->getPlugInName()==outputPlugIn->getPlugInName())
				{
					QTreeWidgetItem* htpcItem = new QTreeWidgetItem(irItem, QStringList(iter->getName()));		
					htpcItem->setData(0, Qt::UserRole, QVariant::fromValue(*iter));
				}				
				++iter;
			}
		}
		++codeIter;
	}	
	// add available output commands
	iter = outputPlugIn->getCommands().constBegin();
	while(iter!=outputPlugIn->getCommands().constEnd())
	{
		// only htpccommands that require an integer value (the ir code) or nothing are valid 
		if (!iter->getData().isValid() || iter->getData().type() == QVariant::Int)
		{
			QListWidgetItem* htpcItem = new QListWidgetItem(iter->getName(), m_outputCommands);
			htpcItem->setData(Qt::UserRole, QVariant::fromValue(*iter));
		}
		++iter;
	}
}

void ConnectionDialog::itemSelected()
{
	QTreeWidgetItem* codeItem = m_irCodeTree->currentItem();
	QListWidgetItem* commandItem = m_outputCommands->currentItem();
	if (codeItem && codeItem->data(0, Qt::UserRole).canConvert<IRCode>())
	{
		m_antiRepeat->blockSignals(true);
		m_antiRepeat->setValue(codeItem->data(0,Qt::UserRole).value<IRCode>().antiRepeat);
		m_antiRepeat->blockSignals(false);
	}
	m_antiRepeat->setEnabled(codeItem!=0);
	if (codeItem && commandItem)
	{
		// allow only associations with IR Codes with HtpcCommands from the output
		m_addCommandButton->setEnabled(
			codeItem->data(0, Qt::UserRole).canConvert<IRCode>() && 
			commandItem->data(Qt::UserRole).canConvert<HtpcCommand>());		
	}
	else
		m_addCommandButton->setEnabled(false);
}

void ConnectionDialog::commandAssociated()
{
	QList<QListWidgetItem*>& htpcItems = m_outputCommands->selectedItems();
	QList<QTreeWidgetItem *>& irItems = m_irCodeTree->selectedItems();
	if (htpcItems.size() > 0 && irItems.size() > 0)
	{
		HtpcCommand command = htpcItems.at(0)->data(Qt::UserRole).value<HtpcCommand>();
		QTreeWidgetItem* item = new QTreeWidgetItem(irItems.at(0), QStringList(command.getName()));
		item->setData(0, Qt::UserRole, QVariant::fromValue(command));
		QList<HtpcCommand>* commandList = 0;
		QMap<IRCode, QList<HtpcCommand>*>::iterator iter = m_commandMap.find(irItems.at(0)->data(0, Qt::UserRole).value<IRCode>());
		if (iter!=m_commandMap.end() && iter.value() != 0)
			commandList = m_commandMap[irItems.at(0)->data(0, Qt::UserRole).value<IRCode>()];
		else
			commandList = new QList<HtpcCommand>;
		commandList->push_back(command);
		m_commandMap[irItems.at(0)->data(0,Qt::UserRole).value<IRCode>()] = commandList;
	}
}

void ConnectionDialog::getCommandMap(QMap<IRCode,QList<HtpcCommand>*>& commandMap)
{
	commandMap = m_commandMap;	
}

void ConnectionDialog::customEvent(QEvent *event)
{
	// if we are in learning mode
	if (m_learning)
	{			
		IRCodeEvent* irEvent = dynamic_cast<IRCodeEvent*>(event);
		// check if the ir event has been learned already 
		if (irEvent && !m_commandMap.contains(irEvent->getCode()))
		{
			// if not create a new ir code
			QTreeWidgetItem* irItem = new QTreeWidgetItem(m_irCodeTree, QStringList(QString::number(irEvent->getCode())));
			IRCode irCode(irEvent->getCode());
			irItem->setData(0, Qt::UserRole, QVariant::fromValue<IRCode>(irCode));							
			m_commandMap[irCode] = 0;
		}
		else if (irEvent)
			QMessageBox::information(this, tr("Error"), tr("IR Code (")+QString::number(irEvent->getCode())+tr(") already in the code tree"));
		m_learning = false;
		m_learnCodeButton->setChecked(false);
	}	
}

void ConnectionDialog::learnButtonPressed()
{
	m_learning = true;
	m_learnCodeButton->setChecked(true);
}

void ConnectionDialog::updateIrCode(int antiRepeat)
{
	QTreeWidgetItem* item = m_irCodeTree->currentItem();
	if (item)
	{
		IRCode& irCode = const_cast<IRCode&>(m_commandMap.find(item->data(0, Qt::UserRole).value<IRCode>()).key());
		irCode.antiRepeat = antiRepeat;		
		item->setData(0, Qt::UserRole, QVariant::fromValue<IRCode>(irCode));
	}
}