/**
Copyright (C) 2006  Volker Wiendl

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License version 2 
as published by the Free Software Foundation.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program; 
if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA
*/

#include "IowPlugIn.h"
#include <QtCore/qplugin.h>
#include <qt/qtimer.h>
#include <qt/qstringlist.h>
#include <qt/qwidget.h>
#include <qt/qfile.h>
#include <qt/qcoreapplication.h>
#include <qt/qapplication.h>
#include <qt/qdatetime.h>

#include "../HtpcControl/MusicInfo.h"
#include "../HtpcControl/SystemInfo.h"

#include "ConnectionDialog.h"
#include "ConfigDialog.h"
#include "IRCodeEvent.h"

char* IowPlugIn::LED_ON("\x00\xF7\xFF");
char* IowPlugIn::LED_OFF("\x00\xFF\xFF");
char* IowPlugIn::ENABLE_RC5  = "\x0c\x01\x00\x00\x00\x00\x00\x00";
char* IowPlugIn::DISABLE_RC5 = "\x0c\x00\x00\x00\x00\x00\x00\x00";
const QString IowPlugIn::m_pluginConfigFile(QApplication::applicationDirPath()+"/iowplugin.cfg");

IowPlugIn::IowPlugIn() : m_run(true),  m_iowHandle(NULL), m_displayBlocked(false), m_displayLineCount(4), 
m_displayLineLength(20), m_scrollStep(0), m_alternateRows(true), m_displayDuration(0)
{	
	// create display timer for continous updates
	m_displayTimer = new QTimer(this);
	connect(m_displayTimer, SIGNAL(timeout()), this, SLOT(displayUpdate()));
	connect(this, SIGNAL(message(const QString&, int, bool)), this, SLOT(showMessage(const QString&, int, bool)));
	// create list of commands supported by this plugin
	// we are able to display text messages
	HtpcCommand command(this, 0, "Text");
	command.setData(QVariant(QVariant::String));
	m_commands.push_back(command);
	// we are able to display music info
	command = HtpcCommand(this, 1, "MusicInfo");	
	command.setData(QVariant::fromValue(MusicInfo()));	
	m_commands.push_back(command);
	// we are able to display system info
	command = HtpcCommand(this, 2, "SystemInfo");
	command.setData(QVariant::fromValue(SystemInfo()));	
	m_commands.push_back(command);
	// try to open IOWarrior
	m_iowHandle = IowKitOpenDevice();
	// if an IO Warrior has been received...
	if(m_iowHandle && IowKitGetProductId(m_iowHandle) == IOWKIT_PID_IOW24)
	{
		IowKitSetWriteTimeout(m_iowHandle, 2000);

		/// enable RC5 special mode
		IowKitWrite( m_iowHandle, 1, ENABLE_RC5, 8 );
		/// switch on LED
		IowKitWrite( m_iowHandle, 0, LED_ON, 3 );
		/// enable IIC Mode
		enableIIC( m_iowHandle, TRUE );
		/// enable LCD special mode
		enableLCD( m_iowHandle, TRUE);	
		// init lcd with 4 lines
		initLCD( m_iowHandle, 4 );	
		// enable display without cursor and blinking
		DisplayControl( m_iowHandle, TRUE, FALSE, FALSE );
		// move forward / shift display
		EntryModeSet( m_iowHandle, TRUE, FALSE );	
	}
	init();
}

void IowPlugIn::init()
{
	if(!m_iowHandle)
		return;	
	// lock the receive method to avoid threading problems
	m_receiveLock.lock();
	QFile file(m_pluginConfigFile);	
	QDomDocument doc("Settings");
	if (doc.setContent(&file))
	{
		// print out the element names of all elements that are direct children
		// of the outermost element.
		QDomElement docElem = doc.documentElement();
		QDomElement settings = docElem.namedItem("IowPlugin").toElement();
		/// enable RC5 special mode
		IowKitWrite( m_iowHandle, 1, settings.attribute("RC5","true").contains("true",Qt::CaseInsensitive) ? ENABLE_RC5 : DISABLE_RC5, 8 );		
		/// switch on LED
		IowKitWrite( m_iowHandle, 0, settings.attribute("LED","true").contains("true",Qt::CaseInsensitive)  ? LED_ON : LED_OFF, 3 );
		/// enable LCD special mode
		enableLCD( m_iowHandle, settings.attribute("LCD","true").contains("true",Qt::CaseInsensitive) ? TRUE : FALSE);	
		QDomElement lcd = settings.namedItem("LCD").toElement();
		m_displayLineLength = lcd.attribute("LineWidth", "20").toInt();
		m_displayLineCount = lcd.attribute("LineCount", "4").toInt();
		m_displayTimer->start(lcd.attribute("DisplayUpdate", "1.0").toFloat() * 1000);
		m_idleMessage = lcd.attribute("IdleMessage");
		m_alternateRows = lcd.attribute("AlternateRowOrder","true").contains("true",Qt::CaseInsensitive);	
		m_musicInfoPattern = lcd.attribute("MusicInfoPattern","%artist\n%time\nTitel: %title - Album: %album").replace("\\n","\n");
	}
	else
		m_displayTimer->start(1000);
	file.close();
	m_receiveLock.unlock();
}

IowPlugIn::~IowPlugIn()
{
	// stop the thread
	stop();
	//cout << "Shutdown"<<endl;
	// delete the qlists 
	QMap<IRCode, QList<HtpcCommand>*>::iterator iter = m_commandMap.begin();
	while (iter!=m_commandMap.end())
	{		
		if (iter.value())
			delete iter.value();
		++iter;
	}
}

/// configure settings for this plugin
void IowPlugIn::configure()
{
	// create ConfigDialog
	ConfigDialog dialog;
	// load current configuration from file
	dialog.loadConfig(m_pluginConfigFile);
	if (dialog.exec()==QDialog::Accepted)
	{
		// write configuration to file
		dialog.saveConfig(m_pluginConfigFile);
		// reload settings
		init();
	}
}

const QString IowPlugIn::getPlugInName() const
{
	return tr("IowPlugIn");
}


void IowPlugIn::stop()
{
	// stop event loop of this thread
	quit();
	// indicate that this plugin is not running anymore
	m_run = false;
	//cout << "Waiting for shutdown"<<endl;
	// terminate if the thread hasn't been stoped already
	terminate();
	// wait until thread is finished
	wait();
	// clear the display
	ClearLCD( m_iowHandle );	
	// disable the LCD
	enableLCD( m_iowHandle, FALSE );
	// disable IIC
	enableIIC( m_iowHandle, FALSE );
	// disable LED
	IowKitWrite( m_iowHandle, 0, LED_OFF, 3 );
	// disable RC5
	IowKitWrite( m_iowHandle, 1, DISABLE_RC5, 8 );
	// close the IOWarrior
	IowKitCloseDevice( m_iowHandle );	
}

bool IowPlugIn::configPlugInConnection(OutputInterface* output, QDomElement& configuration)
{
	// create a config dialog
	ConnectionDialog configDialog;
	// lock command map that no other thread is able to change it
	m_commandMap_lock.lockForRead();
	// init config dialog with the current command map
	configDialog.init(m_commandMap, output);	
	m_commandMap_lock.unlock();
	// display config dialog
	if (configDialog.exec()==QDialog::Accepted)
	{
		// if the user has pressed ok, we replace the current command map with the new one
		m_commandMap_lock.lockForWrite();
		configDialog.getCommandMap(m_commandMap);
		m_commandMap_lock.unlock();
	}
	return true;
}

bool IowPlugIn::connectPlugIn(OutputInterface *output, const QDomElement& configuration)
{
	/// get the available commands for given output interface
	const QList<HtpcCommand>& outputCommands = output->getCommands();	
	/// get the IRCodes for this plugin from the configuration
	const QDomNodeList& irCodes = configuration.elementsByTagName("IRCode");
	for (unsigned int i=0; i<irCodes.length(); ++i)
	{
		bool converted = false;
		// get the code of the IR command
		 IRCode irCode(irCodes.item(i).toElement().attribute("code").toInt(&converted));
		// if the conversion failed, try next code
		if (!converted) continue;
		irCode.antiRepeat = irCodes.item(i).toElement().attribute("antiRepeat", "100").toInt();
		// get the associated HtpcCommands
		const QDomNodeList& htpcCommands = irCodes.item(i).toElement().elementsByTagName("HtpcCommand");
		for (unsigned int j=0; j<htpcCommands.length(); ++j)
		{			
			// get the id of the HtpcCommand
			const int id = htpcCommands.item(j).toElement().attribute("id").toInt(&converted);
			if (!converted)	continue;
			// search the output interface for a corresponding command
			QList<HtpcCommand>::const_iterator iter = outputCommands.constBegin();
			while (iter!=outputCommands.constEnd())
			{			
				// if the output interface command has the same ID as the command in the configuration...
				if ((*iter).getID() == id)
				{
					m_commandMap_lock.lockForWrite();
					QList<HtpcCommand>* commands = 0;
					// if there's already an existing command list for this irCode...
					if (m_commandMap.contains(irCode))
						// chose the existing one
						commands = m_commandMap[irCode];
					else
						// otherwise create a new list
						commands = new QList<HtpcCommand>;				
					// add HtpcCommand to the QList for this ircode
					commands->push_back(*iter);
					// update command map
					m_commandMap[irCode] = commands;
					m_commandMap_lock.unlock();
					// no more HtpcCommands with the same ID should be available in the output interface, so we can break
					break;
				}
				++iter;
			}		
		}
	}
	return true;
}

bool IowPlugIn::receiveCommand(const HtpcCommand& command) 
{
	// create the time line for music info
	const QString TIMELINE = QChar::fromAscii('|') + QString(m_displayLineLength-2, QChar::fromAscii('-')) + QChar::fromAscii('|');
	// lock the receive method to avoid multiple entries by several calls of receiveCommand
	m_receiveLock.lock();
	// get the command id
	switch(command.getID())
	{		
	case 0: // Text Command		
		emit message(command.getData().toString(), 5000);
		break;
	case 1:
		{
			// Music Info 
			// get the music info
			const MusicInfo& info = command.getData().value<MusicInfo>();		
			//printf("Length: %d:%02d\n", length/(1000*60), (length/1000)%60);			
			// create timeline
			QString timeline(TIMELINE);
			// create time string
			QString time = QString::number(info.seconds_played / 60) + ":";
			if (info.seconds_played%60 <10)
				time+="0";
			time+=QString::number(info.seconds_played%60);			
			// check status
			if (info.status==0)			
				time = tr("stopped");
			// if paused, display blinking paused string
			else if (info.status==3 && m_scrollStep%2==0)
				time = "paused";
			// calculate index for displaying the time in the timeline
			int index = (int)( 1.0f + (float)info.seconds_played / (float)info.seconds_overall * (18.0f - time.length() + 1));
			// write time or status into the timeline
			timeline.replace(index, time.length(), time);							
			QString displayString = m_musicInfoPattern;
			displayString.replace("%artist", info.artist);
			displayString.replace("%album", info.album);
			displayString.replace("%title", info.title);
			displayString.replace("%genre", info.genre);
			displayString.replace("%year", QString::number(info.year));
			displayString.replace("%time", timeline);
			// print info
			emit message(displayString, 5000);
		}
		break;
	case 2: // incoming system message
		{			
			// get system message
			const SystemInfo& info = command.getData().value<SystemInfo>();
			// print system message onto the lcd and block the display for duration seconds
			emit message(info.text, info.duration, true);
		}		
	default:
		// unknown htpc command, return false (should result in a delete of the connection)
		m_receiveLock.unlock();
		return false;
	}
	m_receiveLock.unlock();
	return true;
}

void IowPlugIn::showMessage(const QString text, int duration, bool blocking /*=false*/)
{
	if (!m_iowHandle)
		return;
	if (!m_displayBlocked || blocking)
	{
		// print the message
		m_lcdMessage = text;
		m_displayDuration = duration;
		m_displayBlocked = blocking;	
		// do an update of the display immediately when a new message arrives
		displayUpdate();
		m_displayTimer->start(m_displayTimer->interval());
	}
}

void IowPlugIn::displayUpdate()
{
	// increase scroll step to display the next characters of long messages
	m_scrollStep++;
	// if nothing happened after the last message	
	if (m_displayDuration  / m_displayTimer->interval() < 0)
	{
		m_displayBlocked = false;
		if (m_idleMessage.isEmpty())
		{
			// display date and time
			if (m_displayTimer->interval() <=1000 )
				printText(QDateTime::currentDateTime().toString(tr("%1 dddd d. MMMM yyyy \n%3 hh:mm:ss"))+tr(" Uhr"));
			else
				printText(QDateTime::currentDateTime().toString(tr("%1 dddd d. MMMM yyyy \n%3 hh:mm"))+tr(" Uhr"));
		}
		else
			printText(m_idleMessage);
	}
	else
	{		
		m_displayDuration -= m_displayTimer->interval();
		if (m_displayDuration <= 0)
			m_displayBlocked = false;
		printText(m_lcdMessage);
	}
}

void IowPlugIn::run()
{
	char packet[8];
	//IowKitSetTimeout(m_iowHandle, 1000);
	// while the thread is marked to run and the IOWarrior Handle is valid
	while (m_run && m_iowHandle)
	{		
		memset( packet, 0x00, 8 );
		packet[0] = 0x0c;
		// read RC5 command from IOWarrior
		IowKitRead( m_iowHandle, 1, packet, 8 );
		// if packet starts with RC5 flag
		if( packet[0] == 0x0c )
		{			
			// get IRCode from packets
			char command = packet[1] & 0x3f;
			command |= ( (~packet[2]) & 0x40 );
			char address = packet[2] & 0x1f;
			int code = (int)( ( address << 8 ) | command );
			//cout << code << endl;
			// check for connected HtpcCommands for the given ir code
			processIRcommand( code );
		}
		// try to sleep for a moment until we're reading the next command
		this->msleep(10);
	}

}

bool IowPlugIn::saveConfig(QDomDocument& doc)
{
	QMap<QString, QDomElement> output_code_map;
	QMap<QString, QDomElement> code_command_map;
	m_commandMap_lock.lockForRead();
	QMap<IRCode, QList<HtpcCommand>*>::const_iterator codeIter = m_commandMap.constBegin();
	QDomElement inputInterface = doc.createElement("InputInterface");
	inputInterface.setAttribute("name", getPlugInName());
	inputInterface.setAttribute("priority", priority());
	doc.documentElement().appendChild(inputInterface);
	while (codeIter != m_commandMap.constEnd())
	{			
		if (codeIter.value())
		{
			QList<HtpcCommand>::iterator iter = codeIter.value()->begin();
			while(iter!=codeIter.value()->end())
			{		
				QDomElement& outputNode = output_code_map[iter->getPlugIn()->getPlugInName()];			
				if (outputNode.isNull())
				{
					outputNode = doc.createElement("OutputInterface");	
					inputInterface.appendChild(outputNode);
					outputNode.setAttribute("name", iter->getPlugIn()->getPlugInName());
					output_code_map[iter->getPlugIn()->getPlugInName()] = outputNode;
				}
				QDomElement& irNode = code_command_map[QString::number(codeIter.key().code)+outputNode.attribute("name")];
				if (irNode.isNull())
				{
					irNode = doc.createElement("IRCode");			
					outputNode.appendChild(irNode);
					irNode.setAttribute("code", QString::number(codeIter.key().code));
					irNode.setAttribute("antiRepeat", QString::number(codeIter.key().antiRepeat));
					code_command_map[QString::number(codeIter.key().code)+outputNode.attribute("name")] = irNode;
				}						
				QDomElement command = doc.createElement("HtpcCommand");			
				irNode.appendChild(command);
				command.setAttribute("name", iter->getName());
				command.setAttribute("id", QString::number(iter->getID()));					
				++iter;
			}
		}
		++codeIter;
	}	
	m_commandMap_lock.unlock();
	return true;
}


void IowPlugIn::processIRcommand(int code)
{
	m_commandMap_lock.lockForRead();
	// get current window
	QWidget* activeWindow = QApplication::activeWindow();
	// send IRCode Event to active window
	if (activeWindow!=0) QCoreApplication::postEvent(activeWindow, new IRCodeEvent(code));
	// check if there are HtpcCommands associated with the given ir code
	QMap<IRCode, QList<HtpcCommand>*>::iterator codeConnections = m_commandMap.find(code);
	// if code is known
	if (codeConnections!=m_commandMap.end() && codeConnections.key().ready())
	{
		IRCode& irCode = const_cast<IRCode&>(codeConnections.key());
		irCode.lastCall = QTime::currentTime();
		// get associated command list
		QList<HtpcCommand>* commands = codeConnections.value();
		QList<HtpcCommand>::iterator iter = commands->begin();
		while(iter!=commands->end())
		{			
			// if htpc command is interested in ir code
			if (iter->getData().canConvert(QVariant::Int))
				iter->setData(QVariant(code));
			// if the receiver plug in couldn't handle the htpc command...
			if (!iter->getPlugIn()->receiveCommand(*iter))
				// ... delete it from the receiver list
				iter = commands->erase(iter);			
			else
				++iter;
		}
	}
	// code is unknown...
	else if (activeWindow!=0)
		// ...send status bar message to active window 
		QCoreApplication::postEvent(activeWindow, new QStatusTipEvent(tr("Unknown IR Code: ")+QString::number(code)));
	m_commandMap_lock.unlock();
}


void IowPlugIn::printText(const QString& text)
{
	// get different lines from the text
	QStringList lines = text.split(QChar::fromAscii('\n'));
	// remove all lines that can't be displayed due to the size limitations of the display
	if (lines.count() > m_displayLineCount)
		lines.erase(lines.begin()+m_displayLineCount, lines.end());
	// print lines on display	
	int offset = 0;
	// log written lines to clear display lines not overwritten by the message afterwards
	int* linesWritten = new int[m_displayLineCount];
	memset(linesWritten, 0, sizeof(int) * m_displayLineCount);
	for (int i = 0; i<lines.count() && i<m_displayLineCount; ++i)
	{
		int line=i+1+offset;
		// HD44780 controler normaly are controling 4 lines
		if (m_alternateRows && (line%4)==2) line+=1;
		else if (m_alternateRows && (line%4)==3) line-=1;
		linesWritten[line-1] = 1;
		if (text.contains(tr("%%1 ").arg(i+1+offset)) && !text.contains(tr("%%1 ").arg(i+2+offset)))
		{			
			linesWritten[m_alternateRows ? line + 1 : line] = 1;
			printLine(lines[i].remove(QRegExp("%\\d+\\s")), m_displayLineLength*2, line, m_scrollStep);		
			++offset;
		}
		else
			printLine(lines[i].remove(QRegExp("%\\d+\\s")), m_displayLineLength, line, m_scrollStep);
	}
	// clear the lines not overwritten by the current lcd message
	for (int i=0; i<m_displayLineCount; ++i)
	{
		if (linesWritten[i] == 0)
			printLine(tr(" "), m_displayLineLength, i+1);
	}
	delete[] linesWritten;
}


void IowPlugIn::printLine(QString text, int lineLength, int lineNumber, unsigned int scrollStep /*=0*/)
{
	// convert umlaut to display charakter
	text.replace(QChar::fromAscii(''), QChar::fromAscii((char)0x9E));
	text.replace(QChar::fromAscii(''), QChar::fromAscii((char)0x99));
	text.replace(QChar::fromAscii(''), QChar::fromAscii((char)0x83), Qt::CaseInsensitive);
	//info.text.replace(QChar::fromAscii(''), QChar::fromAscii(0xAF));
	text.replace(QChar::fromAscii(''), QChar::fromAscii((char)0x8E), Qt::CaseInsensitive);
	//info.text.replace(QChar::fromAscii(''), QChar::fromAscii(0xB1));

	// message for one line
	QString message;
	// check if the text exceeds the line length
	if (text.length() > lineLength)
	{		
		message = text.right(text.length() - (scrollStep%text.length()));
		message+= " - ";
		message+= text;
		message = message.left(lineLength);
	}
	else
		message = text.leftJustified(lineLength, QChar::fromAscii(' '));
	// go to specified line number
	setDRAMaddressByLineNumber(m_iowHandle, lineNumber );	
	// print out the first m_displayLineLength charakters to display
	LCD_PrintLine(m_iowHandle, message.left(m_displayLineLength).toAscii().constData());
	// if we can use two lines...
	if (lineLength == m_displayLineLength * 2)
	{
		// goto next line
		setDRAMaddressByLineNumber(m_iowHandle, m_alternateRows ? lineNumber + 2 : lineNumber + 1);
		LCD_PrintLine(m_iowHandle, message.right(message.length() - m_displayLineLength).toAscii().constData());
	}
}

// The following methods are from Bjrn Stickler http://intex.ath.cx/io-warrior.aspx?lang=de
// print line with up to 20 characters to lcd
void IowPlugIn::LCD_PrintLine( IOWKIT_HANDLE iowHandle, const char* string )
{
	char flag = (char) 0x80;
	const char* current = string;
	while( *current )
	{
		char packet[8];
		memset( packet, 0x00, 8 );

		packet[0] = 0x05;
		packet[1] = flag | (char)min( strlen(current), 6 );
		memcpy( packet+2, current, (char)min( strlen(current), 6 ) );

		IowKitWrite( iowHandle, 1, packet, 8 );

		current += min( strlen(current), 6 );
	}
}

// set lcd address to write to
long IowPlugIn::setDRAMaddress( IOWKIT_HANDLE iowHandle, long address )
{
	char packet[8];
	memset( packet, 0x00, 8 );
	packet[0] = 0x05;
	packet[1] = 0x01;
	packet[2] = 0x80 | (char)address;
	return IowKitWrite( iowHandle, 1, packet, 8 );	
}

// set lcd address to write to by line number
long IowPlugIn::setDRAMaddressByLineNumber( IOWKIT_HANDLE iowHandle, int lineNumber )
{
	long address = 0x00;
	if( lineNumber < 1 || lineNumber > 4 ) return 0;
	if( lineNumber == 2 ) address |= 0x40;
	else if( lineNumber == 3 ) address |= 0x14;
	else if( lineNumber == 4 ) address |= 0x54;
	return setDRAMaddress( iowHandle, address );
}

// enable iic mode of iowarrior
long IowPlugIn::enableIIC( IOWKIT_HANDLE iowHandle, BOOLEAN enable )
{
	char packet[8];
	memset( packet, 0, 8 );
	packet[0] = 0x01;
	packet[1] = (char)enable;
	return IowKitWrite( iowHandle, 1, packet, 8 );
}

// enable lcd
long IowPlugIn::enableLCD( IOWKIT_HANDLE iowHandle, BOOLEAN enable )
{
	IOWKIT_SPECIAL_REPORT report;
	memset(&report, 0, IOWKIT_SPECIAL_REPORT_SIZE);
	report.ReportID = 0x04; //ReportID for LCD-Mode
	report.Bytes[0] = enable ? 0x01 : 0x00;// enable / disable LCD Special Function
	return IowKitWrite(iowHandle, IOW_PIPE_SPECIAL_MODE, (char *) &report, IOWKIT_SPECIAL_REPORT_SIZE);
}

// initialize lcd
long IowPlugIn::initLCD( IOWKIT_HANDLE iowHandle, long numLines )
{
	IOWKIT_SPECIAL_REPORT report;
	memset(&report, 0, IOWKIT_SPECIAL_REPORT_SIZE);
	report.ReportID = 0x05; //ReportID for LCD-Mode "write"
	report.Bytes[0] = 0x04; // R/S mode=0 / Send bytes
	report.Bytes[1] = 0x38; // function "8-Bit mode"
	report.Bytes[2] = 0x0F; // function "display on"
	report.Bytes[3] = 0x01; // function "clear display"
	report.Bytes[4] = 0x06; // function "entry mode set"				
	//char packet[8] = { 0x05, 0x03, 0x28, 0x01, 0x0c, 0x00, 0x00, 0x00 };
	//packet[2] |= numLines > 1 ? 0x10 : 0x00 ;
	return IowKitWrite(iowHandle, IOW_PIPE_SPECIAL_MODE, (char *) &report, IOWKIT_SPECIAL_REPORT_SIZE);
}

// setup display control 
long IowPlugIn::DisplayControl( IOWKIT_HANDLE iowHandle, BOOLEAN displayOn, BOOLEAN cursorOn, BOOLEAN cursorBlink )
{			
	IOWKIT_SPECIAL_REPORT report;
	memset(&report, 0, IOWKIT_SPECIAL_REPORT_SIZE);
	report.ReportID = 0x05;
	report.Bytes[0] = 0x01;
	report.Bytes[1] = 0x08;
	report.Bytes[1] |= displayOn ? 0x04 : 0x00;
	report.Bytes[1] |= cursorOn ? 0x02 : 0x00;
	report.Bytes[1] |= cursorBlink ? 0x01 : 0x00;
	return IowKitWrite(iowHandle, IOW_PIPE_SPECIAL_MODE, (char *) &report, IOWKIT_SPECIAL_REPORT_SIZE);
}

// setup entrymode
long IowPlugIn::EntryModeSet( IOWKIT_HANDLE iowHandle, BOOLEAN moveForward, BOOLEAN shiftDisplay )
{
	IOWKIT_SPECIAL_REPORT report;
	memset(&report, 0, IOWKIT_SPECIAL_REPORT_SIZE);
	report.ReportID = 0x05;
	report.Bytes[0] = 0x01;
	report.Bytes[1] = 0x04;
	report.Bytes[1] |= moveForward ? 0x02 : 0x00;
	report.Bytes[1] |= shiftDisplay ? 0x01 : 0x00;
	return IowKitWrite(iowHandle, IOW_PIPE_SPECIAL_MODE, (char *) &report, IOWKIT_SPECIAL_REPORT_SIZE);
}

long IowPlugIn::ClearLCD( IOWKIT_HANDLE iowHandle )
{
	IOWKIT_SPECIAL_REPORT report;
	memset(&report, 0, IOWKIT_SPECIAL_REPORT_SIZE);
	report.ReportID = 0x05;
	report.Bytes[0] = 0x01;
	report.Bytes[1] = 0x01;
	return IowKitWrite(iowHandle, IOW_PIPE_SPECIAL_MODE, (char *) &report, IOWKIT_SPECIAL_REPORT_SIZE);
}

Q_EXPORT_STATIC_PLUGIN(IowPlugIn)
Q_EXPORT_PLUGIN2(iowplugin, IowPlugIn)