/**
	Copyright (C) 2006  Volker Wiendl

    This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License version 2 as published 
	by the Free Software Foundation.
    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
	without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
	See the GNU General Public License for more details.
    You should have received a copy of the GNU General Public License along with this program; 
	if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA
*/
#ifndef IOWPLUGIN_H_
#define IOWPLUGIN_H_


#include "../HTPCControl/HtpcCommand.h"
#include "../HTPCControl/InputInterface.h"
#include "../HTPCControl/OutputInterface.h"
#include <qt/qobject.h>
#include <qt/qlist.h>
#include <qt/qmutex.h>
#include <qt/QReadWriteLock.h>
#include <windows.h>

#include "ircode.h"
#include "iowkit.h"
class QTimer;

class IowPlugIn : public InputInterface, public OutputInterface
{
	Q_OBJECT
	Q_INTERFACES(InputInterface OutputInterface)

public:

	/// Constructor
	IowPlugIn();
	/// Destructor
	virtual ~IowPlugIn();

	// HtpcInterface Implementations
	/**
	 * Returns the name of the plugin
	 * @return const QString the name of the plugin
	 */
	const QString getPlugInName() const;
	void configure();

	// InputInterface Implementations
	/// get configuration from QDomElement and connect the output interface with this plugin
	bool connectPlugIn(OutputInterface* output, const QDomElement& configuration);	
	bool configPlugInConnection(OutputInterface* output, QDomElement& configuration);
	bool saveConfig(QDomDocument& doc);
	/// stops the input interface thread
	void stop();
	
	// OutputInterface Implementations
	const QList<HtpcCommand>& getCommands() const {return m_commands;} 
	bool receiveCommand (const HtpcCommand&) ;

signals:
	/** Singal to display message on LCD
	 * @param msg The message to display
	 * @param duration duration to display the message
	 * @param blocking if blocking message set to true
	 */
	void message(const QString& msg, int duration, bool blocking = false);

protected:
	/// the main thread of the input interface
	void run();


private slots:
	
	void displayUpdate();
	void showMessage(const QString text, int duration, bool blocking = false);

private:

	// commands the output interface understands
	QList<HtpcCommand> m_commands;
	// IR Code - Output Command Map
	QMap<IRCode, QList<HtpcCommand>*> m_commandMap;

	/// handles a specific ir code
	void processIRcommand(int code);
	
	/// thread running indicator
	volatile bool m_run;
	/// Mutex for the ircode-htpccommand map
	QReadWriteLock	m_commandMap_lock;
	/// Receive lock to avoid parallel access
	QMutex			m_receiveLock;
	/// the io-warrior handle
	IOWKIT_HANDLE	m_iowHandle;

	/// indicates if a message blocks the display
	bool			m_displayBlocked;

	QTimer*			m_displayTimer;

	void init();
	void printText(const QString& text);
	void printLine(QString text, int lineLength, int lineNumber, unsigned int scrollStep = 0);

	void LCD_PrintLine( IOWKIT_HANDLE iowHandle, const char* string );
	long setDRAMaddress( IOWKIT_HANDLE iowHandle, long address );
	long setDRAMaddressByLineNumber( IOWKIT_HANDLE iowHandle, int lineNumber );
	long enableIIC( IOWKIT_HANDLE iowHandle, BOOLEAN enable );
	long enableLCD( IOWKIT_HANDLE iowHandle, BOOLEAN enable );
	long initLCD( IOWKIT_HANDLE iowHandle, long numLines );
	long DisplayControl( IOWKIT_HANDLE iowHandle, BOOLEAN displayOn, BOOLEAN cursorOn, BOOLEAN cursorBlink );
	long EntryModeSet( IOWKIT_HANDLE iowHandle, BOOLEAN moveForward, BOOLEAN shiftDisplay );
	long ClearLCD( IOWKIT_HANDLE iowHandle );

	/// number of rows of the connected LCD
	int		m_displayLineCount;
	/// number of characters of one line of the connected LCD
	int		m_displayLineLength;
	/// display duration for the last text message
	int		m_displayDuration;
	/// on HD44780 display controllers normaly the rows are addressed in alternating order
	bool	m_alternateRows;
	/// The current lcd message
	QString m_lcdMessage;
	/// idle message
	QString m_idleMessage;
	/// music info apperance
	QString m_musicInfoPattern;

	unsigned int m_scrollStep;

	static char* LED_ON;
	static char* LED_OFF;
	static char* ENABLE_RC5;
	static char* DISABLE_RC5;

	static const QString m_pluginConfigFile;
};


#endif