/**
	Copyright (C) 2006  Volker Wiendl

    This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License version 2 as published 
	by the Free Software Foundation.
    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
	without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
	See the GNU General Public License for more details.
    You should have received a copy of the GNU General Public License along with this program; 
	if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA
*/

#include "VlcPlugIn.h"
#include <QtCore/qplugin.h>
#include <qt/QStringList.h>
#include <windows.h>
#include "../HtpcControl/MusicInfo.h"
#include "../HtpcControl/SystemInfo.h"

#include "ConnectionDialog.h"

#define VLC_PLAYPAUSE	   13000
#define VLC_PAUSE		   13001
#define VLC_PREVIOUSTRACK  13013
#define VLC_NEXTTRACK	   13014
#define VLC_STOP		   13003
#define VLC_SLOWDOWN	   13016
#define VLC_SPEEDUP		   13015
#define VLC_DISPLAYTIME    13004
#define VLC_STEPBACKWARD   13005
#define VLC_STEPFORWARD    13006
#define VLC_VOLUMEUP	   13018
#define VLC_VOLUMEDOWN	   13019
#define VLC_VOLUMEMUTE	   13020
#define VLC_FULLSCREEN	   13017 
#define VLC_ASCPECTRATIO   13027
#define VLC_CUTRATIO	   13028


VlcPlugIn::VlcPlugIn() : m_run(true), m_systemInfoDuration(1000)
{
	m_commands.push_back(HtpcCommand(this, 0, "Play/Pause"));
	m_commands.push_back(HtpcCommand(this, 1, "Pause"));
	m_commands.push_back(HtpcCommand(this, 2, "Stop"));
	m_commands.push_back(HtpcCommand(this, 3, "Previouse Track"));
	m_commands.push_back(HtpcCommand(this, 4, "Next Track"));
	m_commands.push_back(HtpcCommand(this, 5, "Volume Up"));
	m_commands.push_back(HtpcCommand(this, 6, "Volume Down"));
	m_commands.push_back(HtpcCommand(this, 7, "Volume Mute"));
	m_commands.push_back(HtpcCommand(this, 8, "Slower Playback"));
	m_commands.push_back(HtpcCommand(this, 9, "Faster Playback"));
	m_commands.push_back(HtpcCommand(this, 10, "Toggle Fullscreen"));
	m_commands.push_back(HtpcCommand(this, 11, "Aspect Ratio"));
	m_commands.push_back(HtpcCommand(this, 12, "Crop Aspect Ratio"));
	m_commands.push_back(HtpcCommand(this, 13, "Display Timestamp"));
}

VlcPlugIn::~VlcPlugIn()
{
	stop();
	//cout << "Shutdown"<<endl;
}

void VlcPlugIn::configure()
{
}


const QString VlcPlugIn::getPlugInName() const
{
	return tr("VLC Input-/Output-PlugIn");
}


void VlcPlugIn::stop()
{
	quit();
	m_run = false;
	//cout << "Waiting for shutdown"<<endl;
	wait();
}

bool VlcPlugIn::configPlugInConnection(OutputInterface* output, QDomElement& configuration)
{
	// create a config dialog
	ConnectionDialog* configDialog = new ConnectionDialog();
	// lock command map that no other thread is able to change it
	m_listenerLock.lockForRead();
	// init config dialog with the current command map
	configDialog->init(m_musicListeners, m_systemListeners, output);	
	m_listenerLock.unlock();
	// display config dialog
	if (configDialog->exec()==QDialog::Accepted)
	{
		// if the user has pressed ok, we replace the current command map with the new one
		m_listenerLock.lockForWrite();
		configDialog->getMusicListeners(m_musicListeners);
		configDialog->getSystemListeners(m_systemListeners);
		m_listenerLock.unlock();
	}
	delete configDialog;
	configDialog = 0;
	return true;
}

bool VlcPlugIn::connectPlugIn(OutputInterface *output, const QDomElement& configuration)
{
	const QList<HtpcCommand>& outputCommands = output->getCommands();	
	const QDomNodeList& listeners = configuration.elementsByTagName("HtpcCommand");
	for (int i=0; i<listeners.length(); ++i)
	{		
		bool convert = false;
		int id = listeners.item(i).toElement().attribute("id").toInt(&convert);
		if (!convert)
			continue;
		QList<HtpcCommand>::const_iterator iter = outputCommands.constBegin();
		while (iter!=outputCommands.constEnd())
		{				
			if ((*iter).getID() == id)
			{				
				m_listenerLock.lockForWrite();
				if (listeners.item(i).toElement().attribute("connectTo").contains("SystemInfo", Qt::CaseInsensitive))
					m_systemListeners.push_back(*iter);
				else
					m_musicListeners.push_back(*iter);
				m_listenerLock.unlock();
			}
			++iter;
		}
	}
	return true;
}

bool VlcPlugIn::receiveCommand(const HtpcCommand& command) 
{
	m_receiveLock.lock();
#ifdef UNICODE
	HWND hwndVlc= FindWindow(L"wxWindowClassNR",NULL);
#else
	HWND hwndVlc = FindWindow("wxWindowClassNR",NULL);
#endif	
	if (hwndVlc != NULL)
	{
		sendSystemInfo(command.getName());
		switch(command.getID())
		{						
		case 0:
			// play/pause
			SendMessage( hwndVlc, WM_COMMAND, (WPARAM) VLC_PLAYPAUSE , (LPARAM)NULL );
			break;			
		case 1:
			// pause
			SendMessage( hwndVlc, WM_COMMAND, (WPARAM) VLC_PAUSE, (LPARAM)NULL );
			break;			
		case 2:
			// Stop
			SendMessage( hwndVlc, WM_COMMAND, (WPARAM) VLC_STOP, (LPARAM)NULL );
			break;			
		case 3:
			// Previous track
			SendMessage( hwndVlc, WM_COMMAND, (WPARAM) VLC_PREVIOUSTRACK, (LPARAM)NULL );
			break;			
		case 4:
			// next track
			SendMessage( hwndVlc, WM_COMMAND, (WPARAM) VLC_NEXTTRACK, (LPARAM)NULL );
			break;			
		case 5:
			// volume up
			SendMessage( hwndVlc, WM_COMMAND, (WPARAM) VLC_VOLUMEUP, (LPARAM)NULL );
			break;			
		case 6:
			// volume down
			SendMessage( hwndVlc, WM_COMMAND, (WPARAM) VLC_VOLUMEDOWN, (LPARAM)NULL );
			break;			
		case 7:
			// volume mute
			SendMessage( hwndVlc, WM_COMMAND, (WPARAM) VLC_VOLUMEMUTE, (LPARAM)NULL );
			break;			
		case 8:
			// slower playback
			SendMessage( hwndVlc, WM_COMMAND, (WPARAM) VLC_SLOWDOWN, (LPARAM)NULL );
			break;			
		case 9:
			// faster playback
			SendMessage( hwndVlc, WM_COMMAND, (WPARAM) VLC_SPEEDUP, (LPARAM)NULL );
			break;			
		case 10:
			// toggle fullscreen
			SendMessage( hwndVlc, WM_COMMAND, (WPARAM) VLC_FULLSCREEN, (LPARAM)NULL );
			break;			
		case 11:
			// aspect ratio
			SendMessage( hwndVlc, WM_COMMAND, (WPARAM) VLC_ASCPECTRATIO, (LPARAM)NULL );
			break;			
		case 12:
			// crop aspect ratio
			SendMessage( hwndVlc, WM_COMMAND, (WPARAM) VLC_CUTRATIO, (LPARAM)NULL );
			break;			
		case 13:
			// display timestamp
			SendMessage( hwndVlc, WM_COMMAND, (WPARAM) VLC_DISPLAYTIME, (LPARAM)NULL );
			break;			
		default:
			m_receiveLock.unlock();
			return false;
		}
	}
	m_receiveLock.unlock();
	return true;
}

void VlcPlugIn::sendSystemInfo(const QString& message)
{
	SystemInfo sysinfo;
	sysinfo.text = message;
	sysinfo.duration = m_systemInfoDuration;

	m_listenerLock.lockForWrite();
	QList<HtpcCommand>::iterator iter = m_systemListeners.begin();
	while (iter!=m_systemListeners.end())
	{				
		// check if receiver plugin wants to have a SystemInfo Struct
		if (iter->getData().canConvert<SystemInfo>())
		{					
			iter->setData(QVariant::fromValue(sysinfo));	
			iter->getPlugIn()->receiveCommand(*iter);
			++iter;
		}
		// or it can read string messages 
		else if (iter->getData().type() == QVariant::String)
		{
			iter->setData(QVariant(message));
			iter->getPlugIn()->receiveCommand(*iter);
			++iter;
		}							
		else 
			iter = m_systemListeners.erase(iter);		
	}
	m_listenerLock.unlock();
}

void VlcPlugIn::run()
{		
	char* localBuffer = new char[MAX_PATH];	
	MusicInfo musicInfo;
	while (m_run)
	{	
#ifdef UNICODE
		m_hwndVlc = FindWindow(L"VLC Media Player",NULL);
#else
		m_hwndVlc = FindWindow("VLC Media Player",NULL);
#endif	
		if (m_hwndVlc != NULL)
		{	
			{
			//	const QString lastArtist = musicInfo.artist;
			//	const QString lastTitle = musicInfo.title;				
			//	musicInfo.clear();				
			//	// get status
			//	// get time info
			///*	musicInfo.status = (int)SendMessage( m_hwndWinamp, WM_WA_IPC, 0, IPC_ISPLAYING );
			//	musicInfo.seconds_played = (int)SendMessage( m_hwndWinamp, WM_WA_IPC, 0, IPC_GETOUTPUTTIME ) / 1000;
			//	musicInfo.seconds_overall = max( 0, (int)SendMessage( m_hwndWinamp, WM_WA_IPC, 1, IPC_GETOUTPUTTIME ) );*/
			//	m_listenerLock.lockForWrite();
			//	QList<HtpcCommand>::iterator iter = m_musicListeners.begin();
			//	while (iter!=m_musicListeners.end())
			//	{				
			//		// check if receiver plugin wants to have a MusicInfo Struct
			//		if (iter->getData().canConvert<MusicInfo>())
			//		{					
			//			iter->setData(QVariant::fromValue(musicInfo));	
			//			iter->getPlugIn()->receiveCommand(*iter);
			//			++iter;
			//		}
			//		// or it can read string messages 
			//		else if (iter->getData().type() == QVariant::String)
			//		{
			//			// if the artist and title has changed
			//			if (lastArtist!=musicInfo.artist && lastTitle!=musicInfo.title)
			//			{
			//				iter->setData(QVariant(musicInfo.artist+"\n"+musicInfo.title));
			//				iter->getPlugIn()->receiveCommand(*iter);
			//			}
			//			++iter;
			//		}					
			//		// skip system message listeners
			//		else 
			//			iter = m_musicListeners.erase(iter);					
			//	}
			//	m_listenerLock.unlock();
			}			
		}
		this->msleep(500);
	}
	delete[] localBuffer;
}

bool VlcPlugIn::saveConfig(QDomDocument& doc)
{
	QDomElement inputInterface = doc.createElement("InputInterface");
	inputInterface.setAttribute("name", getPlugInName());
	inputInterface.setAttribute("priority", priority());
	doc.documentElement().appendChild(inputInterface);
	QMap<QString, QDomElement> output_map;
	m_listenerLock.lockForRead();
	QList<HtpcCommand>::const_iterator iter = m_musicListeners.constBegin();
	while (iter != m_musicListeners.constEnd())
	{								
		QDomElement& outputNode = output_map[iter->getPlugIn()->getPlugInName()];			
		if (outputNode.isNull())
		{
			outputNode = doc.createElement("OutputInterface");	
			inputInterface.appendChild(outputNode);
			outputNode.setAttribute("name", iter->getPlugIn()->getPlugInName());
			output_map[iter->getPlugIn()->getPlugInName()] = outputNode;
		}
		QDomElement& musicListenerNode = doc.createElement("HtpcCommand");
		outputNode.appendChild(musicListenerNode);
		musicListenerNode.setAttribute("id", QString::number(iter->getID()));
		musicListenerNode.setAttribute("name", iter->getName());
		musicListenerNode.setAttribute("connectTo", "MusicInfo");
		++iter;		
	}	
	iter = m_systemListeners.constBegin();
	while (iter != m_systemListeners.constEnd())
	{								
		QDomElement& outputNode = output_map[iter->getPlugIn()->getPlugInName()];			
		if (outputNode.isNull())
		{
			outputNode = doc.createElement("OutputInterface");	
			inputInterface.appendChild(outputNode);
			outputNode.setAttribute("name", iter->getPlugIn()->getPlugInName());
			output_map[iter->getPlugIn()->getPlugInName()] = outputNode;
		}
		QDomElement& systemListenerNode = doc.createElement("HtpcCommand");
		outputNode.appendChild(systemListenerNode);
		systemListenerNode.setAttribute("id", QString::number(iter->getID()));
		systemListenerNode.setAttribute("name", iter->getName());
		systemListenerNode.setAttribute("connectTo", "SystemInfo");
		++iter;		
	}	
	m_listenerLock.unlock();
	return true;
}

Q_EXPORT_STATIC_PLUGIN(VlcPlugIn)
Q_EXPORT_PLUGIN2(winampplugin, VlcPlugIn)