/**
	Copyright (C) 2006  Volker Wiendl

    This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License version 2 as published 
	by the Free Software Foundation.
    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
	without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
	See the GNU General Public License for more details.
    You should have received a copy of the GNU General Public License along with this program; 
	if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA
*/

#include "WinampPlugIn.h"
#include <QtCore/qplugin.h>
#include <qt/QStringList.h>
#include <windows.h>
#include "../HtpcControl/MusicInfo.h"
#include "../HtpcControl/SystemInfo.h"

#include "ConnectionDialog.h"

#include "winamp/wa_ipc.h"

#define METASIZE 128

WinampPlugIn::WinampPlugIn() : m_run(true), m_systemInfoDuration(1000)
{
	m_commands.push_back(HtpcCommand(this, 0, "Next Track"));
	m_commands.push_back(HtpcCommand(this, 1, "Previous Track"));
	m_commands.push_back(HtpcCommand(this, 2, "Play"));
	m_commands.push_back(HtpcCommand(this, 3, "Pause"));
	m_commands.push_back(HtpcCommand(this, 4, "Stop"));
	m_commands.push_back(HtpcCommand(this, 5, "Fast forward"));
	m_commands.push_back(HtpcCommand(this, 6, "Fast backward"));
	m_commands.push_back(HtpcCommand(this, 7, "Volume up"));
	m_commands.push_back(HtpcCommand(this, 8, "Volume down"));
	m_commands.push_back(HtpcCommand(this, 9, "Load standard playlist"));
}

WinampPlugIn::~WinampPlugIn()
{
	stop();
	//cout << "Shutdown"<<endl;
}

void WinampPlugIn::configure()
{
}


const QString WinampPlugIn::getPlugInName() const
{
	return tr("Winamp Input-/Output-PlugIn");
}


void WinampPlugIn::stop()
{
	quit();
	m_run = false;
	//cout << "Waiting for shutdown"<<endl;
	wait();
}

bool WinampPlugIn::configPlugInConnection(OutputInterface* output, QDomElement& configuration)
{
	// create a config dialog
	ConnectionDialog* configDialog = new ConnectionDialog();
	// lock command map that no other thread is able to change it
	m_listenerLock.lockForRead();
	// init config dialog with the current command map
	configDialog->init(m_musicListeners, m_systemListeners, output);	
	m_listenerLock.unlock();
	// display config dialog
	if (configDialog->exec()==QDialog::Accepted)
	{
		// if the user has pressed ok, we replace the current command map with the new one
		m_listenerLock.lockForWrite();
		configDialog->getMusicListeners(m_musicListeners);
		configDialog->getSystemListeners(m_systemListeners);
		m_listenerLock.unlock();
	}
	delete configDialog;
	configDialog = 0;
	return true;
}

bool WinampPlugIn::connectPlugIn(OutputInterface *output, const QDomElement& configuration)
{
	const QList<HtpcCommand>& outputCommands = output->getCommands();	
	const QDomNodeList& listeners = configuration.elementsByTagName("HtpcCommand");
	for (int i=0; i<listeners.length(); ++i)
	{		
		bool convert = false;
		int id = listeners.item(i).toElement().attribute("id").toInt(&convert);
		if (!convert)
			continue;
		QList<HtpcCommand>::const_iterator iter = outputCommands.constBegin();
		while (iter!=outputCommands.constEnd())
		{				
			if ((*iter).getID() == id)
			{				
				m_listenerLock.lockForWrite();
				if (listeners.item(i).toElement().attribute("connectTo").contains("SystemInfo", Qt::CaseInsensitive))
					m_systemListeners.push_back(*iter);
				else
					m_musicListeners.push_back(*iter);
				m_listenerLock.unlock();
			}
			++iter;
		}
	}
	return true;
}

bool WinampPlugIn::receiveCommand(const HtpcCommand& command) 
{
	m_receiveLock.lock();
#ifdef UNICODE
	HWND hwndWinamp = FindWindow(L"Winamp v1.x",NULL);
#else
	HWND hwndWinamp = FindWindow("Winamp v1.x",NULL);
#endif	
	if (hwndWinamp != NULL)
	{
		sendSystemInfo(command.getName());
		switch(command.getID())
		{						
		case 0:
			// next track
			SendMessage( hwndWinamp, WM_COMMAND, WINAMP_BUTTON5, (LPARAM)NULL );
			break;			
		case 1:
			// previous track
			SendMessage( hwndWinamp, WM_COMMAND, WINAMP_BUTTON1, (LPARAM)NULL );
			break;			
		case 2:
			// play
			SendMessage( hwndWinamp, WM_COMMAND, WINAMP_BUTTON2, (LPARAM)NULL );
			break;			
		case 3:
			// pause
			SendMessage( hwndWinamp, WM_COMMAND, WINAMP_BUTTON3, (LPARAM)NULL );
			break;			
		case 4:
			// stop
			SendMessage( hwndWinamp, WM_COMMAND, WINAMP_BUTTON4, (LPARAM)NULL );
			break;			
		case 5:
			// fast forward
			SendMessage( hwndWinamp, WM_COMMAND, WINAMP_BUTTON5_SHIFT, (LPARAM)NULL );
			break;			
		case 6:
			// fast backward
			SendMessage( hwndWinamp, WM_COMMAND, WINAMP_BUTTON1_SHIFT, (LPARAM)NULL );
			break;			
		case 7:
			// volume up
			SendMessage( hwndWinamp, WM_COMMAND, WINAMP_VOLUMEUP, (LPARAM)NULL );
			break;			
		case 8:
			// volume down
			SendMessage( hwndWinamp, WM_COMMAND, WINAMP_VOLUMEDOWN, (LPARAM)NULL );
			break;
		default:
			m_receiveLock.unlock();
			return false;
		}
	}
	m_receiveLock.unlock();
	return true;
}

void WinampPlugIn::sendSystemInfo(const QString& message)
{
	SystemInfo sysinfo;
	sysinfo.text = message;
	sysinfo.duration = m_systemInfoDuration;

	m_listenerLock.lockForWrite();
	QList<HtpcCommand>::iterator iter = m_systemListeners.begin();
	while (iter!=m_systemListeners.end())
	{				
		// check if receiver plugin wants to have a SystemInfo Struct
		if (iter->getData().canConvert<SystemInfo>())
		{					
			iter->setData(QVariant::fromValue(sysinfo));	
			iter->getPlugIn()->receiveCommand(*iter);
			++iter;
		}
		// or it can read string messages 
		else if (iter->getData().type() == QVariant::String)
		{
			iter->setData(QVariant(message));
			iter->getPlugIn()->receiveCommand(*iter);
			++iter;
		}							
		else 
			iter = m_systemListeners.erase(iter);		
	}
	m_listenerLock.unlock();
}

void WinampPlugIn::run()
{		
	char* localBuffer = new char[MAX_PATH];	
	MusicInfo musicInfo;
	while (m_run)
	{	
#ifdef UNICODE
		m_hwndWinamp = FindWindow(L"Winamp v1.x",NULL);
#else
		m_hwndWinamp = FindWindow("Winamp v1.x",NULL);
#endif	
		if (m_hwndWinamp != NULL)
		{							
			if (readWinampToLocal((char *)SendMessage(m_hwndWinamp, WM_WA_IPC, playingTrack(), IPC_GETPLAYLISTFILE), localBuffer, MAX_PATH)>0)
			{	
				const QString lastArtist = musicInfo.artist;
				const QString lastTitle = musicInfo.title;				
				musicInfo.clear();				
				if (metaInfo("Artist", localBuffer)!=0)
				{
					musicInfo.artist = localBuffer;
					musicInfo.title = metaInfo("Title", localBuffer);
					musicInfo.album = metaInfo("Album", localBuffer);
					musicInfo.length = QString(metaInfo("Length", localBuffer)).toInt();
					musicInfo.year = QString(metaInfo("Year", localBuffer)).toInt();
					musicInfo.bitrate = QString(metaInfo("Bitrate", localBuffer)).toInt();
					//printf("Length: %d:%02d\n", length/(1000*60), (length/1000)%60);
				}
				else
				{
					basicInfo(localBuffer);
					QStringList metaData = QString(localBuffer).split(QChar::fromAscii('-'), QString::SkipEmptyParts);
					switch (metaData.size())
					{
						default: 
							if (metaData.size()==0)
								break;
						case 3:
							musicInfo.album = metaData.at(2);
						case 2: 
							musicInfo.title = metaData.at(1);
						case 1:
							musicInfo.artist = metaData.at(0);
					}					
				}
				// get status
				// get time info
				musicInfo.status = (int)SendMessage( m_hwndWinamp, WM_WA_IPC, 0, IPC_ISPLAYING );
				musicInfo.seconds_played = (int)SendMessage( m_hwndWinamp, WM_WA_IPC, 0, IPC_GETOUTPUTTIME ) / 1000;
				musicInfo.seconds_overall = max( 0, (int)SendMessage( m_hwndWinamp, WM_WA_IPC, 1, IPC_GETOUTPUTTIME ) );
				m_listenerLock.lockForWrite();
				QList<HtpcCommand>::iterator iter = m_musicListeners.begin();
				while (iter!=m_musicListeners.end())
				{				
					// check if receiver plugin wants to have a MusicInfo Struct
					if (iter->getData().canConvert<MusicInfo>())
					{					
						iter->setData(QVariant::fromValue(musicInfo));	
						iter->getPlugIn()->receiveCommand(*iter);
						++iter;
					}
					// or it can read string messages 
					else if (iter->getData().type() == QVariant::String)
					{
						// if the artist and title has changed
						if (lastArtist!=musicInfo.artist && lastTitle!=musicInfo.title)
						{
							iter->setData(QVariant(musicInfo.artist+"\n"+musicInfo.title));
							iter->getPlugIn()->receiveCommand(*iter);
						}
						++iter;
					}					
					// skip system message listeners
					else 
						iter = m_musicListeners.erase(iter);					
				}
				m_listenerLock.unlock();
			}			
		}
		this->msleep(500);
	}
	delete[] localBuffer;
}

bool WinampPlugIn::saveConfig(QDomDocument& doc)
{
	QDomElement inputInterface = doc.createElement("InputInterface");
	inputInterface.setAttribute("name", getPlugInName());
	inputInterface.setAttribute("priority", priority());
	doc.documentElement().appendChild(inputInterface);
	QMap<QString, QDomElement> output_map;
	m_listenerLock.lockForRead();
	QList<HtpcCommand>::const_iterator iter = m_musicListeners.constBegin();
	while (iter != m_musicListeners.constEnd())
	{								
		QDomElement& outputNode = output_map[iter->getPlugIn()->getPlugInName()];			
		if (outputNode.isNull())
		{
			outputNode = doc.createElement("OutputInterface");	
			inputInterface.appendChild(outputNode);
			outputNode.setAttribute("name", iter->getPlugIn()->getPlugInName());
			output_map[iter->getPlugIn()->getPlugInName()] = outputNode;
		}
		QDomElement& musicListenerNode = doc.createElement("HtpcCommand");
		outputNode.appendChild(musicListenerNode);
		musicListenerNode.setAttribute("id", QString::number(iter->getID()));
		musicListenerNode.setAttribute("name", iter->getName());
		musicListenerNode.setAttribute("connectTo", "MusicInfo");
		++iter;		
	}	
	iter = m_systemListeners.constBegin();
	while (iter != m_systemListeners.constEnd())
	{								
		QDomElement& outputNode = output_map[iter->getPlugIn()->getPlugInName()];			
		if (outputNode.isNull())
		{
			outputNode = doc.createElement("OutputInterface");	
			inputInterface.appendChild(outputNode);
			outputNode.setAttribute("name", iter->getPlugIn()->getPlugInName());
			output_map[iter->getPlugIn()->getPlugInName()] = outputNode;
		}
		QDomElement& systemListenerNode = doc.createElement("HtpcCommand");
		outputNode.appendChild(systemListenerNode);
		systemListenerNode.setAttribute("id", QString::number(iter->getID()));
		systemListenerNode.setAttribute("name", iter->getName());
		systemListenerNode.setAttribute("connectTo", "SystemInfo");
		++iter;		
	}	
	m_listenerLock.unlock();
	return true;
}

// Most of these code is copied from the WINAMP Developer Forum

//returns NULL if alloc failed, returns remote address otherwise
void* WinampPlugIn::allocateWinampBuffer(unsigned long bufsize)
{
	void *remoteBuf;
	HANDLE hWinamp;
	unsigned long dWinamp;

	// find the process id
	GetWindowThreadProcessId(m_hwndWinamp, &dWinamp);

	// open the process object
	hWinamp = OpenProcess(PROCESS_ALL_ACCESS,false,dWinamp);
	if(hWinamp == NULL) return NULL;

	// allocate chunk of memory in winamp's address space
	remoteBuf = VirtualAllocEx(hWinamp, NULL, bufsize, MEM_COMMIT, PAGE_READWRITE);

	CloseHandle(hWinamp);
	return remoteBuf;
}

// returns 0 on success (it's winamp's problem if it fails... right?)
int WinampPlugIn::freeWinampBuffer(void *remoteBuf, unsigned long bufsize)
{
	int isError;
	HANDLE hWinamp;
	unsigned long dWinamp;

	// find the process id
	GetWindowThreadProcessId(m_hwndWinamp, &dWinamp);

	// open the process object
	hWinamp = OpenProcess(PROCESS_ALL_ACCESS,false,dWinamp);
	if(hWinamp == NULL) return 2;

	// free the memory in winamp's space
	isError = VirtualFreeEx(hWinamp, remoteBuf, bufsize, MEM_DECOMMIT);
	if(!isError)
	{
		CloseHandle(hWinamp);
		return isError;
	}
	// release it
	isError = VirtualFreeEx(hWinamp, remoteBuf, 0, MEM_RELEASE);
	if(!isError)
	{
		CloseHandle(hWinamp);
		return isError;
	}

	CloseHandle(hWinamp);
	return 0;
}

// returns how many bytes were copied to local buffer
unsigned long WinampPlugIn::readWinampToLocal(void *remoteBuf, void *localBuf, unsigned long bufsize)
{
	int isError;
	HANDLE hWinamp;
	unsigned long dWinamp;

	// find the process id
	GetWindowThreadProcessId(m_hwndWinamp, &dWinamp);

	// open the process object
	hWinamp = OpenProcess(PROCESS_ALL_ACCESS,false,dWinamp);
	if(hWinamp == NULL) return 0;

	isError = ReadProcessMemory(hWinamp, remoteBuf, localBuf, bufsize, NULL);

	CloseHandle(hWinamp);

	if(!isError) return 0;
	else return bufsize;
}

// returns how many bytes were written to winamp's space
unsigned long WinampPlugIn::writeLocalToWinamp(void *localBuf, void *remoteBuf, unsigned long bufsize)
{
	int isError;
	HANDLE hWinamp;
	unsigned long dWinamp;

	// find the process id
	GetWindowThreadProcessId(m_hwndWinamp, &dWinamp);

	// open the process object
	hWinamp = OpenProcess(PROCESS_ALL_ACCESS,false,dWinamp);
	if(hWinamp == NULL) return 0;

	isError = WriteProcessMemory(hWinamp, remoteBuf, localBuf, bufsize, NULL);

	CloseHandle(hWinamp);

	if(!isError) return 0;
	else return bufsize;
}

int WinampPlugIn::playingTrack()
{
	return SendMessage(m_hwndWinamp,WM_WA_IPC,0,IPC_GETLISTPOS);
}

char* WinampPlugIn::metaInfo(char *whichInfo, char *localBuf)
{
	char *remoteBuf, *whichBuf, *structBuf;
	static extendedFileInfoStruct xstruct;
	char* returnVal = localBuf;

	// allocate the memory in winamp's space
	remoteBuf = (char *)allocateWinampBuffer(METASIZE);
	whichBuf = (char *)allocateWinampBuffer(sizeof(whichInfo));
	structBuf = (char *)allocateWinampBuffer(sizeof(xstruct));

	// fill in ext struct
	xstruct.filename = (char *)SendMessage(m_hwndWinamp, WM_WA_IPC, playingTrack(), IPC_GETPLAYLISTFILE);
	xstruct.metadata = whichBuf;
	xstruct.ret = remoteBuf;
	xstruct.retlen = METASIZE;

	// write the local buffers to winamp's space
	writeLocalToWinamp(whichInfo, whichBuf, strlen(whichInfo));
	writeLocalToWinamp(&xstruct, structBuf, sizeof(xstruct));

	// send the "get ext info call"
	if(!SendMessage(m_hwndWinamp,WM_WA_IPC,(int)structBuf,IPC_GET_EXTENDED_FILE_INFO_HOOKABLE))
		returnVal = 0;
	else
		// copy the remote buffer to local space
		readWinampToLocal(remoteBuf, localBuf, METASIZE);		
	
	// free up the memory we allocated in winamp's space
	freeWinampBuffer(remoteBuf, METASIZE);
	freeWinampBuffer(whichBuf, sizeof(whichBuf));
	freeWinampBuffer(structBuf, sizeof(xstruct));
	return returnVal;
}

char* WinampPlugIn::basicInfo(char* localBuf)
{
	char *remoteBuf, *filenameBuf, *titleBuf;
	static basicFileInfoStruct  basicFileInfo;
	char* returnVal = localBuf;

	//basicFileInfoStruct basicFileInfo;

	// allocate the memory in winamp's space
	remoteBuf = (char *)allocateWinampBuffer(sizeof(basicFileInfoStruct));
	filenameBuf = (char *)allocateWinampBuffer(_MAX_FNAME);
	titleBuf = (char *)allocateWinampBuffer(256);
	
	//memset( filename, 0x00, _MAX_FNAME );
	//memset( title, 0x00, 256 );
	memset( &basicFileInfo, 0x00, sizeof( basicFileInfoStruct ) );		
	
	basicFileInfo.filename = filenameBuf;
	basicFileInfo.title = titleBuf;
	basicFileInfo.titlelen = 256;

	// write the local buffers to winamp's space	
	writeLocalToWinamp(&basicFileInfo, remoteBuf, sizeof(basicFileInfoStruct));

	// get file info
	SendMessage( m_hwndWinamp, WM_WA_IPC, (WPARAM)remoteBuf, IPC_GET_BASIC_FILE_INFO );		

	// copy the remote buffer to local space
	readWinampToLocal(titleBuf, localBuf, 256);		
	
	// free up the memory we allocated in winamp's space
	freeWinampBuffer(remoteBuf, sizeof(basicFileInfoStruct));
	freeWinampBuffer(filenameBuf, _MAX_FNAME);
	freeWinampBuffer(titleBuf, 256);
	return returnVal;

}


Q_EXPORT_STATIC_PLUGIN(WinampPlugIn)
Q_EXPORT_PLUGIN2(winampplugin, WinampPlugIn)