#include "ConfigDialog.h"
#include "CommandDialog.h"

#include <qt/qmessagebox.h>
#include <qt/qdom.h>
#include <qt/qfile.h>
#include <qt/qtextstream.h>
#include <qt/qtimer.h>
#include <qt/qcolordialog.h>


ConfigDialog::ConfigDialog(int commands, QWidget* parent /*=0*/, Qt::WFlags wflags /*= 0*/) : QDialog(parent, wflags)
{
	setupUi(this);
	m_commandDlg = new CommandDialog(commands, this);
	m_commandDlg->setVisible(false);
	m_testRect = new QWidget(this, Qt::SplashScreen);
	m_testRect->setWindowOpacity(0.8);
	QHBoxLayout* layout = new QHBoxLayout(m_testRect);
	layout->setMargin(0);
	layout->setSpacing(0);
	m_testTimer = new QTimer(this);
	m_testTimer->setSingleShot(true);	
	connect(m_testTimer, SIGNAL(timeout()), m_testRect, SLOT(hide()));
	m_fontTest = new QLabel("Font Test", m_testRect);
	m_fontTest->setAlignment(Qt::AlignHCenter | Qt::AlignVCenter);
	m_fontTest->setFont(m_osdFont->currentFont());
	m_testRect->layout()->addWidget(m_fontTest);
	m_testRect->setVisible(false);
	connect(m_osdFont, SIGNAL(currentFontChanged(const QFont&)), this, SLOT(updateRect()));
	connect(m_xPos, SIGNAL(valueChanged(int)), this, SLOT(updateRect()));
	connect(m_yPos, SIGNAL(valueChanged(int)), this, SLOT(updateRect()));
	connect(m_width, SIGNAL(valueChanged(int)), this, SLOT(updateRect()));
	connect(m_height, SIGNAL(valueChanged(int)), this, SLOT(updateRect()));
	connect(m_color, SIGNAL(clicked()), this, SLOT(selectColor()));
	connect(m_musicInfoPattern, SIGNAL(textChanged(const QString&)), m_fontTest, SLOT(setText(const QString&)));
	connect(m_fontSize, SIGNAL(valueChanged(int)), this, SLOT(updateRect()));
	connect(m_addCommandButton, SIGNAL(clicked()), this, SLOT(addCommand()));
}

ConfigDialog::~ConfigDialog(void)
{
}

void ConfigDialog::loadConfig(const QString& configFile)
{
	QFile file(configFile);	
	QDomDocument doc("Settings");
	if (doc.setContent(&file))
	{		
		QDomElement docElem = doc.documentElement();
		QDomElement settings = docElem.namedItem("WindowsPlugIn").toElement();
		QDomElement element = settings.namedItem("Mouse").toElement();
		m_mouseAcceleration->setValue(element.attribute("Acceleration","10").toInt());
		element = settings.namedItem("Windows").toElement();
		m_shutdownMessage->setText(element.attribute("ShutdownMessage", tr("Windows wird heruntergefahren")));
		m_shutdownDelay->setValue(element.attribute("ShutdownDelay", "30").toInt());
		element = settings.namedItem("OSD").toElement();
		m_musicInfoPattern->setText(element.attribute("MusicInfoPattern","%artist\\nTitel: %title - Album: %album"));
		m_musicInfoTimeout->setValue(element.attribute("MusicInfoDelay", "2.0").toFloat());
		m_width->blockSignals(true);
		m_width->setValue(element.attribute("width", "640").toInt());
		m_width->blockSignals(false);
		m_height->blockSignals(true);
		m_height->setValue(element.attribute("height","100").toInt());
		m_height->blockSignals(false);
		m_xPos->blockSignals(true);
		m_xPos->setValue(element.attribute("x", "400").toInt());
		m_xPos->blockSignals(false);
		m_yPos->blockSignals(true);
		m_yPos->setValue(element.attribute("y", "800").toInt());
		m_yPos->blockSignals(false);
		m_osdQueue->setChecked(element.attribute("useQueue", "false").contains("true", Qt::CaseInsensitive));
		element = element.namedItem("Font").toElement();
		m_osdFont->setCurrentIndex(m_osdFont->findText(element.attribute("family", "Arial"),Qt::MatchExactly));
		m_fontSize->setValue(element.attribute("size", "20").toInt());
		m_color->setText(element.attribute("color", "#00ff00"));		
		QPalette pal = m_fontTest->palette();
		pal.setColor(QPalette::WindowText, QColor(m_color->text()));
		m_fontTest->setPalette(pal);
		element = settings.namedItem("Commands").toElement();
		QDomNodeList commands = element.elementsByTagName("Command");
		for (int i=0; i<commands.size(); ++i)
		{
			QDomElement command = commands.at(i).toElement();
			m_commandDlg->addCommand(command.attribute("id").toInt(), command.attribute("value"));
		}
	}
	file.close();
}


void ConfigDialog::saveConfig(const QString &configFile)
{	
	QFile file(configFile);	
	if (file.open(QFile::WriteOnly | QFile::Truncate)) 
	{
        QTextStream out(&file);
		QDomDocument doc("Settings");		
		QDomElement rootNode = doc.createElement("Settings");
		doc.appendChild(rootNode);
		QDomElement settings = doc.createElement("WindowsPlugIn");
		QDomElement element = doc.createElement("Mouse");
		element.setAttribute("Acceleration", QString::number(m_mouseAcceleration->value()));
		settings.appendChild(element);
		element = doc.createElement("Windows");		
		element.setAttribute("ShutdownMessage", m_shutdownMessage->text());
		element.setAttribute("ShutdownDelay", m_shutdownDelay->value());
		settings.appendChild(element);
		element = doc.createElement("OSD");
		element.setAttribute("MusicInfoPattern", m_musicInfoPattern->text().replace("\n","\\n"));
		element.setAttribute("MusicInfoDelay", m_musicInfoTimeout->value());
		element.setAttribute("width",m_width->value());
		element.setAttribute("height",m_height->value());
		element.setAttribute("x",m_xPos->value());
		element.setAttribute("y",m_yPos->value());
		element.setAttribute("useQueue", m_osdQueue->checkState()==Qt::Checked ? "true" : "false");
		QDomElement font = doc.createElement("Font");
		font.setAttribute("family", m_osdFont->currentText());
		font.setAttribute("size", m_fontSize->value());
		font.setAttribute("color", m_color->text());
		element.appendChild(font);
		settings.appendChild(element);
		element = doc.createElement("Commands");
		m_commandDlg->saveCommands(element);
		settings.appendChild(element);
		rootNode.appendChild(settings);
		doc.save(out, 6);
		file.close();
	}
	file.close();
}

void ConfigDialog::updateRect()
{
	m_fontTest->setFont(QFont(m_osdFont->currentFont().family(), m_fontSize->value()));	
	m_fontTest->setText(m_fontTest->text().replace("\\n", "\n"));
	m_testRect->resize(m_width->value(), m_height->value());
	m_testRect->move(m_xPos->value(), m_yPos->value());
	m_testRect->setVisible(true);
	
	m_width->setValue(m_testRect->width());
	m_height->setValue(m_testRect->height());
	m_testTimer->start(1000);
}

void ConfigDialog::selectColor()
{
	QColor color = QColorDialog::getColor(Qt::green, this);
	if (color.isValid()) {
		m_color->setText(color.name());
		QPalette pal = m_fontTest->palette();
		pal.setColor(QPalette::WindowText, color);
		m_fontTest->setPalette(pal);
		updateRect();
	}
}

void ConfigDialog::addCommand()
{
	m_commandDlg->setVisible(true);	
}