/**
	Copyright (C) 2006  Volker Wiendl

    This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License version 2 as published 
	by the Free Software Foundation.
    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
	without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
	See the GNU General Public License for more details.
    You should have received a copy of the GNU General Public License along with this program; 
	if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA
*/

#include "WindowsPlugIn.h"
#include "OSD.h"
#include "ConfigDialog.h"
#include "ConnectionDialog.h"
#include <QtCore/qplugin.h>
#include <qt/qstring.h>
#include <qt/qtimer.h>
#include <qt/qprocess.h>
#include <qt/qfile.h>
#include <qt/qcursor.h>
#include <qt/qapplication.h>
#include <qt/qcoreapplication.h>
#include <qt/qevent.h>
#include <math.h>
#include "../HTPCControl/SystemInfo.h"
#include "../HtpcControl/MusicInfo.h"

#include <windows.h>

WindowsPlugIn::WindowsPlugIn() : m_shutdown(false),m_mouseAcceleration(10), m_shutDownDelay(30), m_musicInfoTimeout(2000), m_musicInfoPattern("%artist\n%title"), m_acceleration(1.0),m_pluginConfigFile(QApplication::applicationDirPath()+"/windowsplugin.cfg")
{
	m_commands.push_back(HtpcCommand(this, 0, "Mouse left"));
	m_commands.push_back(HtpcCommand(this, 1, "Mouse right"));
	m_commands.push_back(HtpcCommand(this, 2, "Mouse up"));
	m_commands.push_back(HtpcCommand(this, 3, "Mouse down"));
	m_commands.push_back(HtpcCommand(this, 4, "Left Mouse Button Click"));
	m_commands.push_back(HtpcCommand(this, 5, "Right Mouse Button Click"));
	m_commands.push_back(HtpcCommand(this, 6, "Windows Shutdown"));
	HtpcCommand command(this, 7, "Text");
	command.setData(QVariant(QVariant::String));
	m_commands.push_back(command);
	// we are able to display music info
	command = HtpcCommand(this, 8, "MusicInfo");	
	command.setData(QVariant::fromValue(MusicInfo()));	
	m_commands.push_back(command);
	// we are able to display system info
	command = HtpcCommand(this, 9, "SystemInfo");
	command.setData(QVariant::fromValue(SystemInfo()));	
	m_commands.push_back(command);
	// create on screen display class
	m_osd = new OSD();
	connect(this, SIGNAL(message(const QString&, int)), m_osd, SLOT(showMessage(const QString&, int)));
	m_timer = new QTimer(this);
	m_systemShutDownTimer = new QTimer(this);
	connect(m_timer, SIGNAL(timeout()), this, SLOT(deccelerateMouse()));
	connect(this, SIGNAL(shutDownStarted(int)), m_systemShutDownTimer, SLOT(start(int)));
	connect(this, SIGNAL(shutDownStoped()), m_systemShutDownTimer, SLOT(stop()));
	connect(m_systemShutDownTimer, SIGNAL(timeout()), this, SLOT(shutDownTimeLeft()));
	connect(this, SIGNAL(mouseMoved(int,int)), this, SLOT(moveMouse(int, int)));
	init();
}

WindowsPlugIn::~WindowsPlugIn()
{	
	quit();
	delete m_osd;
}

void WindowsPlugIn::init()
{
	// lock the receive method to avoid threading problems
	m_receiveLock.lock();
	QFile file(m_pluginConfigFile);	
	QDomDocument doc("Settings");
	if (doc.setContent(&file))
	{
		// print out the element names of all elements that are direct children
		// of the outermost element.
		QDomElement docElem = doc.documentElement();
		QDomElement settings = docElem.namedItem("WindowsPlugIn").toElement();
		QDomElement element = settings.namedItem("Mouse").toElement();
		m_mouseAcceleration = element.attribute("Acceleration", "10").toInt();
		element = settings.namedItem("Windows").toElement();
		m_shutDownMessage = element.attribute("ShutdownMessage", tr("Windows wird heruntergefahren"));
		m_shutDownDelay = (element.attribute("ShutdownDelay", "30").toInt() * 1000);
		element = settings.namedItem("OSD").toElement();
		m_musicInfoPattern = element.attribute("MusicInfoPattern","%artist\n%time\nTitel: %title - Album: %album").replace("\\n","\n");
		m_musicInfoTimeout = element.attribute("MusicInfoDelay", "2.0").toFloat() * 1000;
		m_osd->resize(element.attribute("width", "640").toInt(), element.attribute("height","100").toInt());
		m_osd->move(element.attribute("x", "400").toInt(), element.attribute("y", "800").toInt());
		m_osd->useQueue(element.attribute("useQueue", "false").contains("true", Qt::CaseInsensitive));
		element = element.namedItem("Font").toElement();
		m_osd->setFont(element.attribute("family","Arial"));
		m_osd->setFontSize(element.attribute("size","20").toInt());
		m_osd->setColor(element.attribute("color","#00ff00"));
		element = settings.namedItem("Commands").toElement();		
		// remove previous loaded commands
		m_commands.erase(m_commands.begin()+10, m_commands.end());
		QDomNodeList commands = element.elementsByTagName("Command");
		for (int i=0; i<commands.size(); ++i)
		{
			QDomElement command = commands.at(i).toElement();			
			m_commands.push_back(HtpcCommand(this, command.attribute("id").toInt(), command.attribute("value")));
		}
	}
	file.close();
	m_receiveLock.unlock();
}

void WindowsPlugIn::configure()
{
	ConfigDialog dialog(m_commands.size());
	dialog.loadConfig(m_pluginConfigFile);
	if (dialog.exec()==QDialog::Accepted)
	{
		dialog.saveConfig(m_pluginConfigFile);
		init();
	}
}

const QString WindowsPlugIn::getPlugInName() const
{
	return tr("WindowsPlugIn");
}

bool WindowsPlugIn::configPlugInConnection(OutputInterface* output, QDomElement& configuration)
{
	// create a config dialog
	ConnectionDialog configDialog;
	// lock command map that no other thread is able to change it
	m_systemMessageLock.lockForRead();
	// init config dialog with the current command map
	configDialog.init(m_systemMessageListener, output);	
	m_systemMessageLock.unlock();
	// display config dialog
	if (configDialog.exec()==QDialog::Accepted)
	{
		// if the user has pressed ok, we replace the current command map with the new one
		m_systemMessageLock.lockForWrite();
		configDialog.getSystemListeners(m_systemMessageListener);
		m_systemMessageLock.unlock();
	}
	return true;
}

bool WindowsPlugIn::connectPlugIn(OutputInterface *output, const QDomElement& configuration)
{
	const QList<HtpcCommand>& outputCommands = output->getCommands();	
	const QDomNodeList& listeners = configuration.elementsByTagName("HtpcCommand");
	for (int i=0; i<listeners.length(); ++i)
	{		
		bool convert = false;
		int id = listeners.item(i).toElement().attribute("id").toInt(&convert);
		if (!convert)
			continue;
		QList<HtpcCommand>::const_iterator iter = outputCommands.constBegin();
		while (iter!=outputCommands.constEnd())
		{				
			if ((*iter).getID() == id)
			{
				m_systemMessageLock.lockForWrite();
				m_systemMessageListener.push_back(*iter);
				m_systemMessageLock.unlock();
			}
			++iter;
		}
	}
	return true;
}

bool WindowsPlugIn::saveConfig(QDomDocument& doc)
{
	QDomElement inputInterface = doc.createElement("InputInterface");
	inputInterface.setAttribute("name", getPlugInName());
	bool running = isRunning();
	inputInterface.setAttribute("priority", priority());
	doc.documentElement().appendChild(inputInterface);
	QMap<QString, QDomElement> output_map;
	m_systemMessageLock.lockForRead();
	QList<HtpcCommand>::const_iterator iter = m_systemMessageListener.constBegin();
	while (iter  != m_systemMessageListener.constEnd())
	{								
		QDomElement& outputNode = output_map[iter->getPlugIn()->getPlugInName()];			
		if (outputNode.isNull())
		{
			outputNode = doc.createElement("OutputInterface");	
			inputInterface.appendChild(outputNode);
			outputNode.setAttribute("name", iter->getPlugIn()->getPlugInName());
			output_map[iter->getPlugIn()->getPlugInName()] = outputNode;
		}
		QDomElement& systemMessageNode = doc.createElement("HtpcCommand");
		outputNode.appendChild(systemMessageNode);
		systemMessageNode.setAttribute("id", QString::number(iter->getID()));
		systemMessageNode.setAttribute("name", iter->getName());
		++iter;		
	}	
	m_systemMessageLock.unlock();
	return true;
}

void WindowsPlugIn::deccelerateMouse()
{
	m_acceleration /= m_mouseAcceleration;
	//emit message(QString::number(m_acceleration), 1000);
	if (m_acceleration<=1.0)
	{
		m_acceleration = 1.0;
		m_timer->stop();
	}
}

void WindowsPlugIn::moveMouse(int dx, int dy)
{
	QCursor::setPos(QCursor::pos().x() + dx, QCursor::pos().y() + dy);
	//emit message(QString::number(dx)+","+QString::number(dy), 1000);
	m_acceleration += log( m_mouseAcceleration * m_acceleration );
	m_timer->start(250);
}

bool WindowsPlugIn::receiveCommand(const HtpcCommand& command) 
{	
	QString displayText = m_lastDisplayText;
	bool retVal = true;
	int duration = m_musicInfoTimeout;
	m_receiveLock.lock();
	switch(command.getID())
	{
	case 0:
		mouseMoved(- m_acceleration, 0);
		break;
	case 1:
		mouseMoved(m_acceleration, 0);
		break;
	case 2:
		mouseMoved(0, - m_acceleration);
		break;
	case 3:
		mouseMoved(0, m_acceleration);
		break;
	case 4:
		mouse_event(MOUSEEVENTF_LEFTDOWN | MOUSEEVENTF_LEFTUP, QCursor::pos().x(), QCursor::pos().y(), 0,0);
		break;
	case 5:
		mouse_event(MOUSEEVENTF_RIGHTDOWN | MOUSEEVENTF_RIGHTUP, QCursor::pos().x(), QCursor::pos().y(), 0,0);
		break;
	case 6:
		if (m_shutdown)
		{
			AbortSystemShutdown(NULL);
			m_shutdown = false;			
			emit shutDownStoped();
		}
		else
		{
			HANDLE hToken; // handle to process token 
			TOKEN_PRIVILEGES tkp; // pointer to token structure
			OpenProcessToken(GetCurrentProcess(),TOKEN_ADJUST_PRIVILEGES | 
				TOKEN_QUERY, &hToken); // Get the LUID for shutdown privilege.              
			LookupPrivilegeValue(NULL, SE_SHUTDOWN_NAME, &tkp.Privileges[0].Luid); 
			tkp.PrivilegeCount = 1; // one privilege to set
			tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;   
			// Get shutdown privilege for this process. 
			AdjustTokenPrivileges(hToken, FALSE, &tkp, 0, (PTOKEN_PRIVILEGES) NULL, 0);
			// Cannot test the return value of AdjustTokenPrivileges.  
#ifdef _DEBUG
			if (GetLastError() != ERROR_SUCCESS)   
				qDebug("AdjustTokenPrivileges enable failed.");
			//ExitWindowsEx(EWX_POWEROFF, SHTDN_REASON_FLAG_PLANNED);
			else
#endif
			{
				InitiateSystemShutdown(NULL, L"Shutdown initialized", m_shutDownDelay/1000, FALSE, FALSE);
				m_shutdownTimeLeft = m_shutDownDelay;
				emit shutDownStarted(1000);
				m_shutdown = true;
			}
		}
		break;
	case 7: // text message
		{
			displayText = command.getData().toString();
		}
		break;
	case 8: // music info
		{
			// get the music info
			const MusicInfo& info = command.getData().value<MusicInfo>();
			displayText = m_musicInfoPattern;
			displayText.replace("%artist", info.artist);
			displayText.replace("%album", info.album);
			displayText.replace("%title", info.title);
			displayText.replace("%genre", info.genre);
			displayText.replace("%year", QString::number(info.year));
			duration = m_musicInfoTimeout;
		}
		break;
	case 9: // system message
		{
			// get system message
			const SystemInfo& info = command.getData().value<SystemInfo>();
			displayText = info.text;
			duration = info.duration;
		}
		break;
	default:
		{

			int index = command.getName().indexOf(" : ");
			if (index == -1)
				retVal = false;
			else
			{
				QString program = command.getName().left(index);
				QString param = command.getName().right(command.getName().length() - index - 3);
				retVal = QProcess::startDetached(program, QStringList(param));					
			}
		}			
	}
	m_receiveLock.unlock();	
	if (displayText != m_lastDisplayText)
	{
		emit message(displayText, duration);
		m_lastDisplayText = displayText;
	}
	return true;
}

void WindowsPlugIn::sendSystemInfo(const QString& message, int duration)
{
	SystemInfo system_message;
	system_message.text = message;
	system_message.duration = duration;
	m_systemMessageLock.lockForRead();
	QList<HtpcCommand>::iterator iter = m_systemMessageListener.begin();
	while (iter!=m_systemMessageListener.end())
	{				
		if (iter->getData().canConvert<SystemInfo>())
		{					
			iter->setData(QVariant::fromValue(system_message));	
			iter->getPlugIn()->receiveCommand(*iter);
			++iter;
		}
		else if (iter->getData().canConvert(QVariant::String))
		{
			iter->setData(QVariant(message));
			iter->getPlugIn()->receiveCommand(*iter);
			++iter;
		}
		else
			iter = m_systemMessageListener.erase(iter);
		
	}
	m_systemMessageLock.unlock();
}

void WindowsPlugIn::stop()
{
	quit();
	wait();
}

void WindowsPlugIn::run()
{
	exec();
}

void WindowsPlugIn::shutDownTimeLeft()
{
	m_shutdownTimeLeft -= 1000;
	sendSystemInfo(m_shutDownMessage + tr(" %1 s").arg(m_shutdownTimeLeft / 1000), 1100);
}

Q_EXPORT_STATIC_PLUGIN(WindowsPlugIn)
Q_EXPORT_PLUGIN2(windowsplugin, WindowsPlugIn)